#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import os
import sys
import re
import logging
import string
import re
import xml.dom.minidom
import time

import pyutilib.services
import pyutilib.common
import pyutilib.misc
import pyutilib.component.core

from coopr.opt.base import *
from coopr.opt.base.solvers import _extract_version
from coopr.opt.results import *
from coopr.opt.solver import *
from coopr.pyomo.base.blockutil import has_discrete_variables

from six import iteritems, StringIO

try:
    unicode
except:
    basestring = str

class GUROBI(OptSolver):
    """The GUROBI LP/MIP solver
    """

    pyutilib.component.core.alias('gurobi', doc='The GUROBI LP/MIP solver')

    def __new__(cls, *args, **kwds):
        try:
            mode = kwds['solver_io']
            if mode is None:
                mode = 'lp'
            del kwds['solver_io']
        except KeyError:
            mode = 'lp'
        #
        if mode  == 'lp':
            return SolverFactory('_gurobi_shell', **kwds)
        if mode == 'python':
            opt = SolverFactory('_gurobi_direct', **kwds)
            if opt is None:
                logging.getLogger('coopr.plugins').error('Python API for GUROBI is not installed')
                return
            return opt
        #
        if mode == 'os':
            opt = SolverFactory('_ossolver', **kwds)
        elif mode == 'nl':
            opt = SolverFactory('_asl', **kwds)
        else:
            logging.getLogger('coopr.plugins').error('Unknown IO type: %s' % mode)
            return
        opt.set_options('solver=gurobi')
        return opt


class GUROBISHELL(ILMLicensedSystemCallSolver):
    """Shell interface to the GUROBI LP/MIP solver
    """

    pyutilib.component.core.alias('_gurobi_shell',  doc='Shell interface to the GUROBI LP/MIP solver')

    def __init__(self, **kwds):
        #
        # Call base class constructor
        #
        kwds['type'] = 'gurobi'
        ILMLicensedSystemCallSolver.__init__(self, **kwds)

        # NOTE: eventually both of the following attributes should be migrated to a common base class.
        # is the current solve warm-started? a transient data member to communicate state information
        # across the _presolve, _apply_solver, and _postsolve methods.
        self.warm_start_solve = False
        # related to the above, the temporary name of the MST warm-start file (if any).
        self.warm_start_file_name = None

        #
        # Define valid problem formats and associated results formats
        #
        self._valid_problem_formats=[ProblemFormat.cpxlp, ProblemFormat.mps]
        self._valid_result_formats={}
        self._valid_result_formats[ProblemFormat.cpxlp] = [ResultsFormat.soln]
        self._valid_result_formats[ProblemFormat.mps] = [ResultsFormat.soln]
        self.set_problem_format(ProblemFormat.cpxlp)

        # Note: Undefined capabilities default to 'None'
        self._capabilities = pyutilib.misc.Options()
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def _default_results_format(self, prob_format):
        return ResultsFormat.soln

    def warm_start_capable(self):
        return True

    #
    # write a warm-start file in the GUROBI MST format, which is *not* the same as the CPLEX MST format.
    #
    def warm_start(self, instance):

        def _error_labeler(obj):
            raise KeyError(
                "GUROBI Warm Start Error: instance contains an object, %s, "
                "that is not found in the symbol map.  Does the instance "
                "match the provided symbol map?" % ( obj.name, ))

        from coopr.pyomo.base import Var

        mst_file = open(self.warm_start_file_name,'w')

        # for each variable in the symbol_map, add a child to the
        # variables element.  Both continuous and discrete are accepted
        # (and required, depending on other options), according to the
        # CPLEX manual.  Note, this assumes that the symbol_map matches
        # the instance...
        output_index = 0
        for block in instance.all_blocks():
            for variable in block.active_components(Var).itervalues():
                for index in variable:
                    var = variable[index]
                    if (var.stale is True) or (var.fixed is True):
                        continue
                    name = self._symbol_map.getSymbol(var, _error_labeler)
                    mst_file.write("%s %s\n" % (name, str(var.value)))

        mst_file.close()

    # over-ride presolve to extract the warm-start keyword, if specified.
    def _presolve(self, *args, **kwds):

        # create a context in the temporary file manager for 
        # this plugin - is "pop"ed in the _postsolve method.
        pyutilib.services.TempfileManager.push()

        # if the first argument is a string (representing a filename),
        # then we don't have an instance => the solver is being applied
        # to a file.
        self.warm_start_solve = kwds.pop( 'warmstart', False )

        # the input argument can currently be one of two things: an instance or a filename.
        # if a filename is provided and a warm-start is indicated, we go ahead and
        # create the temporary file - assuming that the user has already, via some external
        # mechanism, invoked warm_start() with a instance to create the warm start file.
        if (self.warm_start_solve is True) and (isinstance(args[0],basestring) is True):
            pass # we assume the user knows what they are doing...
        elif (self.warm_start_solve is True) and (isinstance(args[0], basestring) is False) and (has_discrete_variables(args[0]) is True):
           # assign the name of the warm start file *before* calling the base class
           # presolve - the base class method ends up creating the command line,
           # and the warm start file-name is (obviously) needed there.
           self.warm_start_file_name = pyutilib.services.TempfileManager.create_tempfile(suffix = '.gurobi.mst')
        else:
           self.warm_start_file_name = None

        # let the base class handle any remaining keywords/actions.
        ILMLicensedSystemCallSolver._presolve(self, *args, **kwds)

        # NB: we must let the base class presolve run first so that the
        # symbol_map is actually constructed!

        if (len(args) > 0) and (isinstance(args[0],basestring) is False):

            # write the warm-start file - currently only supports MIPs.
            # we only know how to deal with a single problem instance.
            if self.warm_start_solve is True:

                if len(args) != 1:
                    raise ValueError(
                        "GUROBI _presolve method can only handle a single "
                        "problem instance - %s were supplied" % (len(args),))

                if has_discrete_variables(args[0]) is True:
                    start_time = time.time()
                    self.warm_start(args[0])
                    end_time = time.time()
                    if self._report_timing is True:
                        print("Warm start write time=%.2f seconds" % (end_time-start_time))

    def executable(self):

        if sys.platform == 'win32':
            executable = pyutilib.services.registered_executable("gurobi.bat")
        else:
            executable = pyutilib.services.registered_executable("gurobi.sh")
        if executable is None:
            pyutilib.component.core.PluginGlobals.env().log.warning("Could not locate the 'gurobi' executable, which is required for solver %s" % self.name)
            self.enable = False
            return None
        return executable.get_path()

    def version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        solver_exec = self.executable()
        if solver_exec is None:
            return _extract_version('')
        outname = pyutilib.services.TempfileManager.create_tempfile(suffix = '.gurobi.version')
        with open(outname,'w') as f:
            # **Note, adding a 'timelimit' keyword here results in empty output for some reason
            results = pyutilib.subprocess.run( [solver_exec], stdin='from gurobipy import *; print gurobi.version(); exit()', ostream=f) 
        tmp = None
        with open(outname,'r') as f:
            tmp = tuple(eval(f.read().strip()))
        while(len(tmp) < 4):
            tmp += (0,)
        return tmp[:4]

    def create_command_line(self,executable,problem_files):

        #
        # Define log file
        # The log file in CPLEX contains the solution trace, but the solver status can be found in the solution file.
        #
        self.log_file = pyutilib.services.TempfileManager.create_tempfile(suffix = '.gurobi.log')

        #
        # Define solution file
        # As indicated above, contains (in XML) both the solution and solver status.
        #
        self.soln_file = pyutilib.services.TempfileManager.create_tempfile(suffix = '.gurobi.txt')
        self.results_file = self.soln_file

        #
        # Write the GUROBI execution script
        #

        problem_filename = self._problem_files[0]
        solution_filename = self.soln_file
        warmstart_filename = self.warm_start_file_name
        if sys.platform == 'win32':
            problem_filename  = problem_filename.replace('\\', r'\\')
            solution_filename = solution_filename.replace('\\', r'\\')
            if self.warm_start is True:
                warmstart_filename = warmstart_filename.replace('\\', r'\\')

        # translate the options into a normal python dictionary, from a
        # pyutilib.component.config.options.SectionWrapper - the
        # gurobi_run function doesn't know about coopr, so the translation
        # is necessary.
        options_dict = {}
        for key in self.options:
            options_dict[key] = self.options[key]

        # NOTE: the gurobi shell is independent of Coopr python virtualized environment, so any
        #       imports - specifically that required to get GUROBI_RUN - must be handled explicitly.
        # NOTE: The gurobi plugin (GUROBI.py) and GUROBI_RUN.py live in the same directory.
        script  = "import sys\n"
        script += "from gurobipy import *\n"
        script += "sys.path.append('%s')\n" % os.path.dirname(__file__)
        script += "from GUROBI_RUN import *\n"
        script += "gurobi_run%s\n" % str((problem_filename, warmstart_filename, solution_filename, self.options.mipgap, options_dict, self.suffixes))
        script += "quit()\n"

        # dump the script and warm-start file names for the
        # user if we're keeping files around.
        if self.keepfiles:
            script_fname = pyutilib.services.TempfileManager.create_tempfile(suffix = '.gurobi.script')
            script_file = open(script_fname, 'w')
            script_file.write( script )
            script_file.close()

            print("Solver script file: '%s'" % script_fname)
            if (self.warm_start_solve is True) and (self.warm_start_file_name is not None):
                print("Solver warm-start file: " + self.warm_start_file_name)

        #
        # Define command line
        #
        cmd = [self.executable()]
        if self._timer:
            cmd.insert(0, self._timer)
        return pyutilib.misc.Bunch( cmd=cmd, script=script, 
                                    log_file=self.log_file, env=None )

    def process_logfile(self):

        return ILMLicensedSystemCallSolver.process_logfile(self)

    def process_soln_file(self,results):

        # check for existence of the solution file
        # not sure why we just return - would think that we
        # would want to indicate some sort of error
        if not os.path.exists(self.soln_file):
            return

        soln = Solution()

        soln_variables = soln.variable
        num_variables_read = 0
        soln_constraints = soln.constraint

        # string compares are too expensive, so simply introduce some section IDs.
        # 0 - unknown
        # 1 - problem
        # 2 - solution
        # 3 - solver

        section = 0 # unknown

        solution_seen = False
        
        range_duals = {}
        range_slacks = {}
        INPUT = open(self.soln_file,"r")
        for line in INPUT:
            line = line.strip()
            tokens = [token.strip() for token in line.split(":")]
            if (tokens[0] == 'section'):
                if (tokens[1] == 'problem'):
                    section = 1
                elif (tokens[1] == 'solution'):
                    section = 2
                    solution_seen = True
                elif (tokens[1] == 'solver'):
                    section = 3
            else:
                if (section == 2):
                    if (tokens[0] == 'var'):
                        if tokens[1] != "ONE_VAR_CONSTANT":
                            soln_variables[tokens[1]] = {"Value" : float(tokens[2]), "Id" : num_variables_read}
                            num_variables_read += 1
                    elif (tokens[0] == 'status'):
                        soln.status = getattr(SolutionStatus, tokens[1])
                    elif (tokens[0] == 'gap'):
                        soln.gap = float(tokens[1])
                    elif (tokens[0] == 'objective'):
                        soln.objective['__default_objective__'].value=float(tokens[1])
                        if results.problem.sense == ProblemSense.minimize:
                            results.problem.upper_bound = float(tokens[1])
                        else:
                            results.problem.lower_bound = float(tokens[1])
                    elif (tokens[0] == 'constraintdual'):
                        name = tokens[1]
                        if name != "c_e_ONE_VAR_CONSTANT":
                            if name.startswith('c_'):
                                soln_constraints[tokens[1]].dual = float(tokens[2])
                            elif name.startswith('r_l_'):
                                range_duals.setdefault(name[4:],[0,0])[0] = float(tokens[2])
                            elif name.startswith('r_u_'):
                                range_duals.setdefault(name[4:],[0,0])[1] = float(tokens[2])
                    elif (tokens[0] == 'constraintslack'):
                        name = tokens[1]
                        if name != "c_e_ONE_VAR_CONSTANT":
                            if name.startswith('c_'):
                                soln_constraints[tokens[1]].slack = float(tokens[2])
                            elif name.startswith('r_l_'):
                                range_slacks.setdefault(name[4:],[0,0])[0] = float(tokens[2])
                            elif name.startswith('r_u_'):
                                range_slacks.setdefault(name[4:],[0,0])[1] = float(tokens[2])
                    elif (tokens[0] == 'varrc'):
                        if tokens[1] != "ONE_VAR_CONSTANT":
                            soln_variables[tokens[1]]["Rc"] = float(tokens[2])
                    else:
                        setattr(soln, tokens[0], tokens[1])
                elif (section == 1):
                    if tokens[0] == 'sense':
                        if tokens[1] == 'minimize':
                            results.problem.sense = ProblemSense.minimize
                        elif tokens[1] == 'maximize':
                            results.problem.sense = ProblemSense.maximize
                    else:
                        try:
                            val = eval(tokens[1])
                        except:
                            val = tokens[1]
                        setattr(results.problem, tokens[0], val)
                elif (section == 3):
                    try:
                        val = eval(tokens[1])
                    except:
                        val = tokens[1]
                    if (tokens[0] == 'status'):
                        results.solver.status = getattr(SolverStatus, val)
                    elif (tokens[0] == 'termination_condition'):
                        try:
                            results.solver.termination_condition = getattr(TerminationCondition, val)
                        except AttributeError:
                            results.solver.termination_condition = TerminationCondition.unknown
                    else:
                        setattr(results.solver, tokens[0], val)

        INPUT.close()

        # For the range constraints, supply only the dual with the largest
        # magnitude (at least one should always be numericly zero)
        scon = soln.Constraint
        for key,(ld,ud) in iteritems(range_duals):
            if abs(ld) > abs(ud):
                scon['r_l_'+key].dual = ld
            else:
                scon['r_u_'+key].dual = ud
        # slacks
        for key,(ls,us) in iteritems(range_slacks):
            if abs(ls) > abs(us):
                scon['r_l_'+key].slack = ls
            else:
                scon['r_u_'+key].slack = us


        if solution_seen is True:
            results.solution.insert(soln)

    def _postsolve(self):

        # take care of the annoying GUROBI log file in the current directory.
        # this approach doesn't seem overly efficient, but python os module functions 
        # doesn't accept regular expression directly.
        filename_list = os.listdir(".")
        for filename in filename_list:
            # IMPT: trap the possible exception raised by the file not existing.
            #       this can occur in pyro environments where > 1 workers are
            #       running GUROBI, and were started from the same directory.
            #       these logs don't matter anyway (we redirect everything),
            #       and are largely an annoyance.
            try:
                if  re.match('gurobi\.log', filename) != None:
                    os.remove(filename)
            except OSError:
                pass

        # let the base class deal with returning results.
        results = ILMLicensedSystemCallSolver._postsolve(self)

        # finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin. does not
        # include, for example, the execution script. but does include 
        # the warm-start file. 
        pyutilib.services.TempfileManager.pop(remove=not self.keepfiles)

        return results

if sys.platform == 'win32':
    pyutilib.services.register_executable(name='gurobi.bat')
else:
    pyutilib.services.register_executable(name='gurobi.sh')



#
# Unit Tests for coopr.plugins.solvers.ASL
#
#

import os
import sys
import fileinput
import subprocess
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
cooprdir = dirname(abspath(__file__))+os.sep+".."+os.sep+".."+os.sep
currdir = dirname(abspath(__file__))+os.sep

from nose.tools import nottest
import pyutilib.th as unittest
import pyutilib.services
import pyutilib.common
import coopr.plugins.solvers
import coopr.opt
import coopr
from coopr.pyomo import *

try:
    ipopt = coopr.plugins.solvers.ASL(keepfiles=True, options={'solver':'ipopt'})
    if (ipopt.executable() is not None) and (ipopt.available(False) is True):
        ipopt_available = True
    else:
        ipopt_available = False
except pyutilib.common.ApplicationError:
    ipopt_available=False

class test_ipopt(unittest.TestCase):

    def setUp(self):
        self.do_setup()

    def do_setup(self):
        global tmpdir
        tmpdir = os.getcwd()
        os.chdir(currdir)
        pyutilib.services.TempfileManager.sequential_files(0)
        pyutilib.services.TempfileManager.tempdir = currdir

        self.asl = coopr.plugins.solvers.ASL(keepfiles=True, options={'solver':'ipopt'})
        self.asl.suffixes=['.*']

        #ver = subprocess.Popen(['ipopt','-v'],stdout=subprocess.PIPE).communicate()[0].split()[1]
        #if ver != '3.9.3':
        #    for line in fileinput.input(currdir+"test_solve_from_nl.baseline",inplace=1):
        #            print line.replace('3.9.3',ver),
        #    for line in fileinput.input(currdir+"test_solve_from_instance.baseline",inplace=1):
        #            print line.replace('3.9.3',ver),

        # The sisser CUTEr instance
        # Formulated in Pyomo by Carl D. Laird, Daniel P. Word, Brandon C. Barrera and Saumyajyoti Chaudhuri
        # Taken from:

        # AMPL Model by Hande Y. Benson
        #
        # Copyright (C) 2001 Princeton University
        # All Rights Reserved
        #
        # Permission to use, copy, modify, and distribute this software and
        # its documentation for any purpose and without fee is hereby
        # granted, provided that the above copyright notice appear in all
        # copies and that the copyright notice and this
        # permission notice appear in all supporting documentation.

        #   Source:
        #   F.S. Sisser,
        #   "Elimination of bounds in optimization problems by transforming
        #   variables",
        #   Mathematical Programming 20:110-121, 1981.

        #   See also Buckley#216 (p. 91)

        #   SIF input: Ph. Toint, Dec 1989.

        #   classification OUR2-AN-2-0

        sisser_instance = ConcreteModel()

        sisser_instance.N = RangeSet(1,2)
        sisser_instance.xinit = Param(sisser_instance.N, initialize={ 1 : 1.0, 2 : 0.1})

        def fa(model, i):
            return value(model.xinit[i])
        sisser_instance.x = Var(sisser_instance.N,initialize=fa)

        def f(model):
            return 3*model.x[1]**4 - 2*(model.x[1]*model.x[2])**2 + 3*model.x[2]**4
        sisser_instance.f = Objective(rule=f,sense=minimize)

        # need to flag variables as used!!!
        sisser_instance.preprocess()

        self.sisser_instance = sisser_instance

    def tearDown(self):
        global tmpdir
        pyutilib.services.TempfileManager.clear_tempfiles()
        os.chdir(tmpdir)
        pyutilib.services.TempfileManager.unique_files()

    @unittest.skipIf(not ipopt_available, "The 'ipopt' executable is not available")
    def test_solve_from_nl(self):
        """ Test ipopt solve from nl file """
        results = self.asl.solve(currdir+"sisser.pyomo.nl", logfile=currdir+"test_solve_from_nl.log")
        # We don't want the test to care about which Ipopt version we are using
        results.Solution(0).Message = "Ipopt"
        results.Solver.Message = "Ipopt"
        results.write(filename=currdir+"test_solve_from_nl.txt", times=False, format='json')
        self.assertMatchesJsonBaseline(currdir+"test_solve_from_nl.txt", currdir+"test_solve_from_nl.baseline", tolerance=1e-7)
        os.remove(currdir+"test_solve_from_nl.log")

    @unittest.skipIf(not ipopt_available, "The 'ipopt' executable is not available")
    def test_solve_from_instance(self):
        """ Test ipopt solve from a pyomo instance and load the solution """
        results = self.asl.solve(self.sisser_instance)
        # We don't want the test to care about which Ipopt version we are using
        results.Solution(0).Message = "Ipopt"
        results.Solver.Message = "Ipopt"
        results.write(filename=currdir+"test_solve_from_instance.txt", times=False, format='json')
        self.assertMatchesJsonBaseline(currdir+"test_solve_from_instance.txt", currdir+"test_solve_from_instance.baseline", tolerance=1e-7)
        self.sisser_instance.load(results)

if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2011 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


__all__ = ["PHSolverServerAction"]

try:
    import cPickle as pickle
except ImportError:
    import pickle
try:
    import Pyro.core
    import pyutilib.pyro
    using_pyro=True
except ImportError:
    using_pyro=False
import time

import pyutilib.misc
import pyutilib.component.core
from pyutilib.enum import Enum

from coopr.opt.parallel.manager import *
from coopr.opt.parallel.async_solver import *
from coopr.opt.results import SolverResults


# 
# an enumerated type used to define specific actions for a PH solver server.
#

PHSolverServerAction = Enum(
   'solve' # perform a scenario solve using whatever data is available
)

#
# a specialized asynchronous solver manager for Progressive Hedging.
#

class SolverManager_PHPyro(AsynchronousSolverManager):

    pyutilib.component.core.alias('phpyro')

    def clear(self):
        """
        Clear manager state
        """
        AsynchronousSolverManager.clear(self)
        self.client = pyutilib.pyro.Client()
        self._verbose = False
        self._ah = {}

    def _perform_queue(self, ah, *args, **kwds):
        """
        Perform the queue operation.  This method returns the ActionHandle,
        and the ActionHandle status indicates whether the queue was successful.
        """

        # the PH solver server expects no non-keyword arguments. 
        if len(args) > 0:
           raise RuntimeError("ERROR: The _perform_queue method of PH pyro solver manager received position input arguments, but accepts none.")

        if "action" not in kwds:
           raise RuntimeError("ERROR: No 'action' keyword supplied to _perform_queue method of PH pyro solver manager")

        if "name" not in kwds:
           raise RuntimeError("ERROR: No 'name' keyword supplied to _perform_queue method of PH pyro solver manager")
        instance_name = kwds["name"]

        if "verbose" in kwds:
            self._verbose = kwds["verbose"]
        else:
            # we always want to pass a verbose flag to the solver server.
            kwds["verbose"] = False 

        #
        # Pickl everything into one big data object via the "Bunch" command and post the task.
        #
        data=pyutilib.misc.Bunch(**kwds)

        # NOTE: the task type (type=) should be the name of the scenario/bundle!

        task = pyutilib.pyro.Task(data=data, id=ah.id)
        self.client.add_task(task, verbose=self._verbose, override_type=instance_name)
        self._ah[task.id] = ah

        return ah

    def _perform_wait_any(self):
        """
        Perform the wait_any operation.  This method returns an
        ActionHandle with the results of waiting.  If None is returned
        then the ActionManager assumes that it can call this method again.
        Note that an ActionHandle can be returned with a dummy value,
        to indicate an error.
        """

        # TBD: ONLY REALLY NEED TO RETURN ANY ONE - MODIFY DISPATCHER TO RETURN THE FIRST OF ANY QUEUE THAT HAS SOMETHING.
        queues_to_check = self.client.queues_with_results()
        if len(queues_to_check) > 0:

           # this protects us against the case where we get an 
           # action handle that we didn't know about or expect.

           for queue_name in queues_to_check:

               task = self.client.get_result(override_type=queue_name)
               if task.id in self._ah:
                   ah = self._ah[task.id]
                   self._ah[task.id] = None
                   ah.status = ActionStatus.done
                   self.results[ah.id] = pickle.loads(task.result)
                   return ah
               else:
                   # TBD: WE SHOULD OBVIOUSLY DO SOMETHING ELSE HERE.
                   print("TROUBLE - GOT RESULTS WE DIDN'T EXPECT!!!")
        else:
          # if the queues are all empty, wait some time for things to fill up.
          # constantly pinging dispatch servers wastes their time, and inhibits
          # task worker communication. the good thing about queues_to_check
          # is that it simultaneously grabs information for any queues with
          # results => one client query can yield many results.
          
          # TBD: We really need to parameterize the time-out value, but it
          #      isn't clear how to propagate this though the solver manager
          #      interface layers.
          time.sleep(0.01)

if not using_pyro:
    SolverManagerFactory.deactivate('phpyro')



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


__all__ = []

try:
    import cPickle as pickle
except ImportError:
    import pickle
try:
    import Pyro.core
    import pyutilib.pyro
    using_pyro=True
except ImportError:
    using_pyro=False

import pyutilib.misc
import pyutilib.component.core

from coopr.opt.parallel.manager import *
from coopr.opt.parallel.async_solver import *
from coopr.opt.results import SolverResults


class SolverManager_Pyro(AsynchronousSolverManager):

    pyutilib.component.core.alias('pyro')

    def clear(self):
        """
        Clear manager state
        """
        AsynchronousSolverManager.clear(self)
        self.client = pyutilib.pyro.Client()
        self._opt = None
        self._verbose = False
        self._ah = {} # maps task ids to their corresponding action handle.
        self._symbol_map = {} # maps task ids to the corresponding symbol map. 

    def _perform_queue(self, ah, *args, **kwds):
        """
        Perform the queue operation.  This method returns the ActionHandle,
        and the ActionHandle status indicates whether the queue was successful.
        """

        if 'opt' in kwds:
            self._opt = kwds['opt']
            del kwds['opt']
        else:
            raise ActionManagerError("No solver passed to SolverManager_Pyro, method=_perform_queue; use keyword option \"opt\"")

        if 'verbose' in kwds:
            self._verbose = kwds['verbose']
            del kwds['verbose']

        #
        # Force coopr.opt to ignore tests for availability, at least locally.
        #
        kwds['available'] = True
        self._opt._presolve(*args, **kwds)
        problem_file_string = open(self._opt._problem_files[0],'r').read()

        #
        # Delete this option, to ensure that the remote worker does the check for
        # availability.
        #
        del kwds['available']

        #
        # We can't pickle the options object itself - so extract a simple
        # dictionary of solver options and re-construct it on the other end.
        #
        solver_options = {}
        for key in self._opt.options:
            solver_options[key]=self._opt.options[key]

        # NOTE: let the distributed node deal with the warm-start
        # pick up the warm-start file, if available.
        warm_start_file_string = None
        warm_start_file_name = None
        if hasattr(self._opt,  "warm_start_solve"):
            if (self._opt.warm_start_solve is True) and (self._opt.warm_start_file_name is not None):
                warm_start_file_name = self._opt.warm_start_file_name
                warm_start_file_string = open(warm_start_file_name, 'r').read()

        #
        # Pickle everything into one big data object via the "Bunch" command
        # and post the task!
        #
        data=pyutilib.misc.Bunch(opt=self._opt.type, \
                                 file=problem_file_string, filename=self._opt._problem_files[0], \
                                 warmstart_file=warm_start_file_string, warmstart_filename=warm_start_file_name, \
                                 kwds=kwds, solver_options=solver_options, suffixes=self._opt.suffixes)

        task = pyutilib.pyro.Task(data=data, id=ah.id)
        self.client.add_task(task, verbose=self._verbose)
        self._ah[task.id] = ah
        self._symbol_map[task.id] = self._opt._symbol_map

        return ah

    def _perform_wait_any(self):
        """
        Perform the wait_any operation.  This method returns an
        ActionHandle with the results of waiting.  If None is returned
        then the ActionManager assumes that it can call this method again.
        Note that an ActionHandle can be returned with a dummy value,
        to indicate an error.
        """
        if self.client.num_results() > 0:
            # this protects us against the case where we get an action
            # handle that we didn't know about or expect.
            while(True):
                task = self.client.get_result()
                if task.id in self._ah:

                    ah = self._ah[task.id]
                    del self._ah[task.id]

                    symbol_map = self._symbol_map[task.id]
                    del self._symbol_map[task.id]

                    ah.status = ActionStatus.done

                    self.results[ah.id] = pickle.loads(task.result)

                    # symbol maps don't pass across the Pyro interface (they
                    # are not pickle-able), so tag the results object with
                    # the cached symbol map.
                    self.results[ah.id]._symbol_map = symbol_map

                    return ah

if not using_pyro:
    SolverManagerFactory.deactivate('pyro')



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['AmplMIPConverter']

from coopr.opt.base import *

from pyutilib.component.core import *
from pyutilib.component.config import *
from pyutilib.component.executables import *
import pyutilib.subprocess
import pyutilib.common


class AmplMIPConverter(ManagedSingletonPlugin):

    implements(IProblemConverter)

    executable = ExtensionPoint(IExternalExecutable)

    def __init__(self,**kwds):
        ManagedSingletonPlugin.__init__(self,**kwds)

    def can_convert(self, from_type, to_type):
        """Returns true if this object supports the specified conversion"""
        #
        # Test if the ampl executable is available
        #
        if self.executable.service("ampl") is None:
            return False
        #
        # Return True for specific from/to pairs
        #
        if from_type == ProblemFormat.mod and to_type == ProblemFormat.nl:
            return True
        if from_type == ProblemFormat.mod and to_type == ProblemFormat.mps:
            return True
        return False

    def apply(self, *args, **kwargs):
        """Convert an instance of one type into another"""
        if not isinstance(args[2],basestring):
            raise ConverterError("Can only apply ampl to convert file data")
        cmd = self.executable.service("ampl").get_path()
        if cmd is None:
            raise ConverterError("The 'ampl' executable cannot be found")
        script_filename = pyutilib.services.TempfileManager.create_tempfile(suffix = '.ampl')
        if args[1] == ProblemFormat.nl:
            output_filename = pyutilib.services.TempfileManager.create_tempfile(suffix = '.nl')
        else:
            output_filename = pyutilib.services.TempfileManager.create_tempfile(suffix = '.mps')
        script_filename = "tmp.ampl"
        cmd += " " + script_filename
        #
        # Create the AMPL script
        #
        OUTPUT = open(script_filename, 'w')
        OUTPUT.write("#\n")
        OUTPUT.write("# AMPL script for converting the following files\n")
        OUTPUT.write("#\n")
        if len(args[2:]) == 1:
            OUTPUT.write('model '+args[2]+";\n")
        else:
            OUTPUT.write('model '+args[2]+";\n")
            OUTPUT.write('data '+args[3]+";\n")
        abs_ofile = os.path.abspath(output_filename)
        if args[1] == ProblemFormat.nl:
            OUTPUT.write('write b'+abs_ofile[:-3]+";\n")
        else:
            OUTPUT.write('write m'+abs_ofile[:-4]+";\n")
        OUTPUT.close()
        #
        # Execute command and cleanup
        #
        output = pyutilib.subprocess.run(cmd)
        if not os.path.exists(output_filename):       #pragma:nocover
            raise pyutilib.common.ApplicationError("Problem launching 'ampl' to create '%s': %s" % (output_filename, output))
        return (output_filename,),None # empty variable map



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2010 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import logging
import re
import itertools

from six import itervalues, iteritems, advance_iterator

_gurobi_version = None
try:
    # import all the glp_* functions
    from gurobipy import *
    # create a version tuple of length 4
    _gurobi_version = gurobi.version()
    while(len(_gurobi_version) < 4):
        _gurobi_version += (0,)
    _gurobi_version = _gurobi_version[:4]
    _GUROBI_VERSION_MAJOR = _gurobi_version[0]
    gurobi_python_api_exists = True
except ImportError:
    gurobi_python_api_exists = False

from pyutilib.misc import Bunch, Options
from pyutilib.component.core import alias
from pyutilib.services import TempfileManager

from coopr.opt.base import *
from coopr.opt.base.solvers import _extract_version
from coopr.opt.results import *
from coopr.opt.solver import *
from coopr.pyomo.base import SymbolMap, NumericLabeler, Suffix
from coopr.pyomo.base.numvalue import value
from coopr.pyomo.base.block import active_subcomponents_generator, active_subcomponents_data_generator

logger = logging.getLogger('coopr.plugins')

GRB_MAX = -1
GRB_MIN = 1

class ModelSOS(object):
    def __init__(self):
        self.sosType = {}
        self.sosName = {}
        self.varnames = {}
        self.weights = {}
        self.block_cntr = 0

    def count_constraint(self,symbol_map,labeler,gurobi_var_map,con,name,level,index=None):

        self.block_cntr += 1
        self.varnames[self.block_cntr] = []
        self.weights[self.block_cntr] = []
        if level == 1:
            self.sosType[self.block_cntr] = GRB.SOS_TYPE1
        elif level == 2:
            self.sosType[self.block_cntr] = GRB.SOS_TYPE2

        # The variable being indexed
        var = con.sos_vars()

        if index is None:
            tmpSet = con.sos_set()
            self.sosName[self.block_cntr] = name
        else:
            tmpSet = con.sos_set()[index]
            self.sosName[self.block_cntr] = name+str(index)

        # Get all the variables
        cntr = 1.0
        for idx in tmpSet:
            self.varnames[self.block_cntr].append(gurobi_var_map[symbol_map.getSymbol(var[idx],labeler)])
            # We need to weight each variable
            # For now we just increment a counter
            self.weights[self.block_cntr].append(cntr)
            cntr += 1.0

class gurobi_direct ( OptSolver ):
    """The Gurobi optimization solver (direct API plugin)

 The gurobi_direct plugin offers an API interface to Gurobi.  It requires the
 Python Gurobi API interface (gurobipy) be in Coopr's lib/ directory.  Generally, if you can run Coopr's Python instance, and execute

 >>> import gurobipy
 >>>

 with no errors, then this plugin will be enabled.

 Because of the direct connection with the Gurobi, no temporary files need be
 written or read.  That ostensibly makes this a faster plugin than the file-based
 Gurobi plugin.  However, you will likely not notice any speed up unless you are
 using the GLPK solver with PySP problems (due to the rapid re-solves).

 One downside to the lack of temporary files, is that there is no LP file to
 inspect for clues while debugging a model.  For that, use the 'write' solver
 option:

 $ pyomo model.{py,dat} \
   --solver=gurobi_direct \
   --solver-options  write=/path/to/some/file.lp

 This is a direct interface to Gurobi's Model.write function, the extension of the file is important.  You could, for example, write the file in MPS format:

 $ pyomo model.{py,dat} \
   --solver=gurobi_direct \
   --solver-options  write=/path/to/some/file.mps

    """

    alias('_gurobi_direct', doc='Direct Python interface to the Gurobi optimization solver.')

    def __init__(self, **kwds):
        #
        # Call base class constructor
        #
        kwds['type'] = 'gurobi_direct'
        OptSolver.__init__(self, **kwds)

        self._model = None

        # NOTE: eventually both of the following attributes should be migrated
        # to a common base class.  Is the current solve warm-started?  A
        # transient data member to communicate state information across the
        # _presolve, _apply_solver, and _postsolve methods.
        self.warm_start_solve = False

        # Note: Undefined capabilities default to 'None'
        self._capabilities = Options()
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def version(self):
        if _gurobi_version is None:
            return _extract_version('')
        return _gurobi_version

    def available(self, exception_flag=True):
        """ True if the solver is available """

        if exception_flag is False:
            return gurobi_python_api_exists
        else:
            if gurobi_python_api_exists is False:
                raise pyutilib.common.ApplicationError("No Gurobi <-> Python bindings available - Gurobi direct solver functionality is not available")
            else:
                return True        

    def _populate_gurobi_instance ( self, model ):

        from coopr.pyomo.base import Var, Objective, Constraint, ConstraintList, IntegerSet, BooleanSet, SOSConstraint
        from coopr.pyomo.expr import canonical_is_constant
        from coopr.pyomo import LinearCanonicalRepn

        try:
            grbmodel = Model( name=model.name )
        except Exception:
            e = sys.exc_info()[1]
            msg = 'Unable to create Gurobi model.  Have you installed the Python'\
            '\n       bindings for Gurobi?\n\n\tError message: %s'
            raise Exception(msg % e)

        labeler = NumericLabeler('x')
        self._symbol_map = SymbolMap(model)

        pyomo_gurobi_variable_map = {} # maps labels to the corresponding Gurobi variable object
        
        for block in model.all_blocks():
            for var_value in active_subcomponents_data_generator(block,Var):
                if (not var_value.active) or (var_value.fixed is True):
                    continue
    
                lb = -GRB.INFINITY
                ub = GRB.INFINITY
    
                if var_value.lb is not None:
                    lb = value(var_value.lb)
                if var_value.ub is not None:
                    ub = value(var_value.ub)
    
                var_value_label = self._symbol_map.getSymbol(var_value, labeler)
                
                # be sure to impart the integer and binary nature of any variables
                if isinstance(var_value.domain, IntegerSet):
                    var_type = GRB.INTEGER
                elif isinstance(var_value.domain, BooleanSet):
                    var_type = GRB.BINARY
                else:
                    var_type = GRB.CONTINUOUS
    
                pyomo_gurobi_variable_map[var_value_label] = grbmodel.addVar(lb=lb, \
                                                                             ub=ub, \
                                                                             vtype=var_type, \
                                                                             name=var_value_label)

        grbmodel.update() 

        model_repn_suffix = model.subcomponent("repn")
        if (model_repn_suffix is None) or (not model_repn_suffix.type() is Suffix) or (not model_repn_suffix.active is True):
            raise ValueError("Unable to find an active Suffix with name 'repn' on block: %s" % (model.cname(True)))

        for obj_data in active_subcomponents_data_generator(model,Objective):
            sense = GRB_MAX
            if obj_data.is_minimizing(): sense = GRB_MIN
            grbmodel.ModelSense = sense
            obj_expr = LinExpr()

            obj_repn = model_repn_suffix.getValue(obj_data)

            if (isinstance(obj_repn, LinearCanonicalRepn) and (obj_repn.linear == None)) or canonical_is_constant(obj_repn):
                msg = "Ignoring objective '%s[%s]' which is constant"
                logger.warning( msg % (str(objective), str(key)) )
                continue

            if isinstance(obj_repn, LinearCanonicalRepn):

                if obj_repn.constant != None:
                    obj_expr.addConstant(obj_repn.constant)

                if obj_repn.linear != None:

                    for i in xrange(len(obj_repn.linear)):
                        var_coefficient = obj_repn.linear[i]
                        var_value = obj_repn.variables[i]                    

                        label = self._symbol_map.getSymbol(var_value, labeler)
                        obj_expr.addTerms(var_coefficient, pyomo_gurobi_variable_map[label])
                        # the coefficients are attached to the model when creating the
                        # variables, below

            else:

                if 0 in obj_repn: # constant term
                    obj_expr.addConstant(obj_repn[0][None])

                if 1 in obj_repn: # first-order terms
                    hash_to_variable_map = obj_repn[-1]
                    for var_hash, var_coefficient in iteritems(obj_repn[1]):
                        label = self._symbol_map.getSymbol(hash_to_variable_map[var_hash], labeler)
                        obj_expr.addTerms(var_coefficient, pyomo_gurobi_variable_map[label])
                        # the coefficients are attached to the model when creating the
                        # variables, below

                if 2 in obj_repn:
                    obj_expr = QuadExpr(obj_expr)
                    hash_to_variable_map = obj_repn[-1]
                    for quad_repn, coef in iteritems(obj_repn[2]):
                        gurobi_expr = QuadExpr(coef)
                        for var_hash, exponent in iteritems(quad_repn):
                            gurobi_var = pyomo_gurobi_variable_map[self._symbol_map.getSymbol(hash_to_variable_map[var_hash], labeler)]
                            gurobi_expr *= gurobi_var 
                            if exponent == 2:
                                gurobi_expr *= gurobi_var
                        obj_expr += gurobi_expr
            # need to cache the objective label, because the GUROBI python interface doesn't track this.
            self._objective_label = self._symbol_map.getSymbol(obj_data, labeler)

        grbmodel.setObjective(obj_expr,sense=sense)       
 
        # SOS constraints - largely taken from cpxlp.py so updates there,
        # should be applied here
        # TODO: Allow users to specify the variables coefficients for custom
        # branching/set orders - refer to cpxlp.py
        sos1 = self._capabilities.sos1
        sos2 = self._capabilities.sos2
        
        modelSOS = ModelSOS()
        for block in model.all_blocks():
            for con in active_subcomponents_generator(block,SOSConstraint):
                level = con.sos_level()
                if (level == 1 and not sos1) or (level == 2 and not sos2) or (level > 2):
                    raise Exception("Solver does not support SOS level %s constraints" % (level,))
                name = self._symbol_map.getSymbol( con, labeler )
                masterIndex = con.sos_set_set()
                if None in masterIndex:
                    # A single constraint
                    modelSOS.count_constraint(self._symbol_map,labeler,pyomo_gurobi_variable_map,con, name, level)
                else:
                    # A series of indexed constraints
                    for index in masterIndex:
                        modelSOS.count_constraint(self._symbol_map,labeler,pyomo_gurobi_variable_map,con, name, level, index)
        
        if modelSOS.sosType:
            for key in modelSOS.sosType:
                grbmodel.addSOS(modelSOS.sosType[key], \
                                modelSOS.varnames[key], \
                                modelSOS.weights[key] )
        
        grbmodel.update()

        # Track the range constraints and there associated variables added by gurobi
        self._last_native_var_idx = grbmodel.NumVars-1
        range_var_idx = grbmodel.NumVars
        _self_range_con_var_pairs = self._range_con_var_pairs = []

        for block in model.all_blocks():
            
            block_repn_suffix = block.subcomponent("repn")
            if (block_repn_suffix is None) or (not block_repn_suffix.type() is Suffix) or (not block_repn_suffix.active is True):
                block_repn_suffix = None

            for constraint in active_subcomponents_generator(block,Constraint):
                if constraint.trivial: 
                    continue
    
                for constraint_data in itervalues(constraint):
    
                    if not constraint_data.active: 
                        continue
                    elif constraint_data.lower is None and constraint_data.upper is None:
                        continue  # not binding at all, don't bother
    
                    con_repn = None
                    if block_repn_suffix is not None:
                        con_repn = block_repn_suffix.getValue(constraint_data)

                    offset = 0.0
                    constraint_label = self._symbol_map.getSymbol(constraint_data, labeler)

                    if isinstance(con_repn, LinearCanonicalRepn):

                        if con_repn.constant != None:
                            offset = con_repn.constant
                        expr = LinExpr() + offset

                        if con_repn.linear != None:
    
                            linear_coefs = list()
                            linear_vars = list()

                            for i in xrange(len(con_repn.linear)):

                                var_coefficient = con_repn.linear[i]
                                var_value = con_repn.variables[i]
    
                                label = self._symbol_map.getSymbol(var_value, labeler)
                                linear_coefs.append( var_coefficient )
                                linear_vars.append( pyomo_gurobi_variable_map[label] )
    
                            expr += LinExpr(linear_coefs, linear_vars)

                    else:                        

                        if 0 in con_repn:
                            offset = con_repn[0][None]
                        expr = LinExpr() + offset

                        if 1 in con_repn: # first-order terms
    
                            linear_coefs = list()
                            linear_vars = list()
    
                            hash_to_variable_map = con_repn[-1]
                            for var_hash, var_coefficient in iteritems(con_repn[1]):
                                var = hash_to_variable_map[var_hash]
                                label = self._symbol_map.getSymbol(var, labeler)
                                linear_coefs.append( var_coefficient )
                                linear_vars.append( pyomo_gurobi_variable_map[label] )
    
                            expr += LinExpr(linear_coefs, linear_vars)

                        if 2 in con_repn: # quadratic constraint
                            if _GUROBI_VERSION_MAJOR < 5:
                                raise ValueError("The gurobi_direct plugin does not handle quadratic constraint expressions\
                            \nfor Gurobi major versions < 5. Current version: Gurobi %s.%s%s" % gurobi.version())
                            expr = QuadExpr(expr)
                            hash_to_variable_map = con_repn[-1]
                            for quad_repn, coef in iteritems(con_repn[2]):
                                gurobi_expr = QuadExpr(coef)
                                for var_hash, exponent in iteritems(quad_repn):
                                    gurobi_var = pyomo_gurobi_variable_map[self._symbol_map.getSymbol(hash_to_variable_map[var_hash], labeler)]
                                    gurobi_expr *= gurobi_var 
                                    if exponent == 2:
                                        gurobi_expr *= gurobi_var
                                expr += gurobi_expr

                    if constraint_data._equality:
                        sense = GRB.EQUAL    # Fixed
                        bound = constraint_data.lower()
                        grbmodel.addConstr(
                            lhs=expr, sense=sense, rhs=bound, name=constraint_label )
                    else:
                        # L <= body <= U
                        if (constraint_data.upper is not None) and (constraint_data.lower is not None):
                            grb_con = grbmodel.addRange(expr, constraint_data.lower(), constraint_data.upper(), constraint_label)
                            _self_range_con_var_pairs.append((grb_con,range_var_idx))
                            range_var_idx += 1
                        # body <= U
                        elif constraint_data.upper is not None:
                            bound = constraint_data.upper()
                            if bound < float('inf'):
                                grbmodel.addConstr(
                                    lhs=expr,
                                    sense=GRB.LESS_EQUAL,
                                    rhs=bound,
                                    name=constraint_label
                                    )
                        # L <= body            
                        else:
                            bound = constraint_data.lower()
                            if bound > -float('inf'):
                                grbmodel.addConstr(
                                    lhs=expr,
                                    sense=GRB.GREATER_EQUAL,
                                    rhs=bound,
                                    name=constraint_label
                                    )
         
        grbmodel.update()

        self._gurobi_instance = grbmodel

    def warm_start_capable(self):
        msg = "Gurobi has the ability to use warmstart solutions.  However, it "\
              "has not yet been implemented into the Coopr gurobi_direct plugin."
        logger.info( msg )
        return False


    def warm_start(self, instance):
        pass


    def _presolve(self, *args, **kwargs):
        from coopr.pyomo.base.PyomoModel import Model

        self.suffixes = kwargs.pop( 'suffixes', self.suffixes )
        self.warm_start_solve = kwargs.pop( 'warmstart', False )
        self.keepfiles = kwargs.pop( 'keepfiles' , False )
        self.tee = kwargs.pop( 'tee', False )

        model = args[0]
        if len(args) != 1:
            msg = "The gurobi_direct plugin method '_presolve' must be supplied "\
                  "a single problem instance - %s were supplied"
            raise ValueError(msg % len(args))
        elif not isinstance(model, Model):
            raise ValueError("The problem instance supplied to the "            \
                 "gurobi_direct plugin '_presolve' method must be of type 'Model'")

        self._populate_gurobi_instance( model )
        grbmodel = self._gurobi_instance

        if 'write' in self.options:
            fname = self.options.write
            grbmodel.write( fname )

        # Scaffolding in place
        if self.warm_start_solve is True:

            if len(args) != 1:
                msg = "The gurobi_direct _presolve method can only handle a single"\
                      "problem instance - %s were supplied"
                raise ValueError(msg % len(args))

            self.warm_start( model )


    def _apply_solver(self):
        # TODO apply appropriate user-specified parameters

        prob = self._gurobi_instance

        if self.tee:
            prob.setParam( 'OutputFlag', self.tee )
        else:
            prob.setParam( 'OutputFlag', 0)
        
        if self.keepfiles == True:
            log_file = TempfileManager.create_tempfile(suffix = '.gurobi.log')
            print("Solver log file: " + log_file)
            prob.setParam('LogFile', log_file)

        #Options accepted by gurobi (case insensitive):
        #['Cutoff', 'IterationLimit', 'NodeLimit', 'SolutionLimit', 'TimeLimit',
        # 'FeasibilityTol', 'IntFeasTol', 'MarkowitzTol', 'MIPGap', 'MIPGapAbs',
        # 'OptimalityTol', 'PSDTol', 'Method', 'PerturbValue', 'ObjScale', 'ScaleFlag',
        # 'SimplexPricing', 'Quad', 'NormAdjust', 'BarIterLimit', 'BarConvTol',
        # 'BarCorrectors', 'BarOrder', 'Crossover', 'CrossoverBasis', 'BranchDir',
        # 'Heuristics', 'MinRelNodes', 'MIPFocus', 'NodefileStart', 'NodefileDir',
        # 'NodeMethod', 'PumpPasses', 'RINS', 'SolutionNumber', 'SubMIPNodes', 'Symmetry',
        # 'VarBranch', 'Cuts', 'CutPasses', 'CliqueCuts', 'CoverCuts', 'CutAggPasses',
        # 'FlowCoverCuts', 'FlowPathCuts', 'GomoryPasses', 'GUBCoverCuts', 'ImpliedCuts',
        # 'MIPSepCuts', 'MIRCuts', 'NetworkCuts', 'SubMIPCuts', 'ZeroHalfCuts', 'ModKCuts',
        # 'Aggregate', 'AggFill', 'PreDual', 'DisplayInterval', 'IISMethod', 'InfUnbdInfo',
        # 'LogFile', 'PreCrush', 'PreDepRow', 'PreMIQPMethod', 'PrePasses', 'Presolve',
        # 'ResultFile', 'ImproveStartTime', 'ImproveStartGap', 'Threads', 'Dummy', 'OutputFlag']
        for key in self.options:
            prob.setParam( key, self.options[key] )
            
        if 'relax_integrality' in self.options:
            for v in prob.getVars():
                if v.vType != GRB.CONTINUOUS:
                    v.vType = GRB.CONTINUOUS
            prob.update()

        if _GUROBI_VERSION_MAJOR >= 5:
            for suffix in self.suffixes:
                if re.match(suffix,"dual"):
                    prob.setParam(GRB.Param.QCPDual,1)

        # Actually solve the problem.
        prob.optimize()

        # FIXME: can we get a return code indicating if Gurobi had a
        # significant failure?
        return Bunch(rc=None, log=None)


    def _gurobi_get_solution_status ( self ):
        status = self._gurobi_instance.Status
        if   GRB.OPTIMAL         == status: return SolutionStatus.optimal
        elif GRB.INFEASIBLE      == status: return SolutionStatus.infeasible
        elif GRB.CUTOFF          == status: return SolutionStatus.other
        elif GRB.INF_OR_UNBD     == status: return SolutionStatus.other
        elif GRB.INTERRUPTED     == status: return SolutionStatus.other
        elif GRB.LOADED          == status: return SolutionStatus.other
        elif GRB.SUBOPTIMAL      == status: return SolutionStatus.other
        elif GRB.UNBOUNDED       == status: return SolutionStatus.other
        elif GRB.ITERATION_LIMIT == status: return SolutionStatus.stoppedByLimit
        elif GRB.NODE_LIMIT      == status: return SolutionStatus.stoppedByLimit
        elif GRB.SOLUTION_LIMIT  == status: return SolutionStatus.stoppedByLimit
        elif GRB.TIME_LIMIT      == status: return SolutionStatus.stoppedByLimit
        elif GRB.NUMERIC         == status: return SolutionStatus.error
        raise RuntimeError('Unknown solution status returned by Gurobi solver')


    def _postsolve(self):

        # the only suffixes that we extract from GUROBI are
        # constraint duals, constraint slacks, and variable
        # reduced-costs. scan through the solver suffix list
        # and throw an exception if the user has specified
        # any others.
        extract_duals = False
        extract_slacks = False
        extract_reduced_costs = False
        for suffix in self.suffixes:
            flag=False
            if re.match(suffix,"dual"):
                extract_duals = True
                flag=True
            if re.match(suffix,"slack"):
                extract_slacks = True
                flag=True
            if re.match(suffix,"rc"):
                extract_reduced_costs = True
                flag=True
            if not flag:
                raise RuntimeError("***The gurobi_direct solver plugin cannot extract solution suffix="+suffix)

        gprob = self._gurobi_instance

        if (gprob.getAttr(GRB.Attr.IsMIP)):
            extract_reduced_costs = False
            extract_duals = False

        pvars = gprob.getVars()
        cons = gprob.getConstrs()
        qcons = []
        if _GUROBI_VERSION_MAJOR >= 5:
            qcons = gprob.getQConstrs()
        
        results = SolverResults()
        soln = Solution()
        problem = results.problem
        solver  = results.solver

        solver.name = "Gurobi %s.%s%s" % gurobi.version()
        # solver.memory_used =
        # solver.user_time = None
        # solver.system_time = None
        solver.wallclock_time = gprob.Runtime
        # solver.termination_condition = None
        # solver.termination_message = None

        problem.name = gprob.ModelName
        problem.lower_bound = None
        problem.upper_bound = None
        problem.number_of_constraints          = len(cons)+len(qcons)+gprob.NumSOS
        problem.number_of_nonzeros             = gprob.NumNZs
        problem.number_of_variables            = gprob.NumVars
        problem.number_of_binary_variables     = gprob.NumBinVars
        problem.number_of_integer_variables    = gprob.NumIntVars
        problem.number_of_continuous_variables = gprob.NumVars \
                                                - gprob.NumIntVars \
                                                - gprob.NumBinVars
        problem.number_of_objectives = 1
        problem.number_of_solutions = gprob.SolCount

        problem.sense = ProblemSense.minimize
        if problem.sense == GRB_MAX: problem.sense = ProblemSense.maximize

        soln.status = self._gurobi_get_solution_status()

        if soln.status in (SolutionStatus.optimal, SolutionStatus.stoppedByLimit):
            obj_val = gprob.ObjVal
            if problem.sense == ProblemSense.minimize:
                problem.lower_bound = obj_val
                if problem.number_of_binary_variables + problem.number_of_integer_variables == 0:
                    problem.upper_bound = obj_val
            else:
                problem.upper_bound = obj_val
                if problem.number_of_binary_variables + problem.number_of_integer_variables == 0:
                    problem.lower_bound = obj_val

            soln.objective[self._objective_label].value = obj_val
            
            # Those variables not added by gurobi due to range constraints
            for var in itertools.islice(pvars,self._last_native_var_idx+1):
                soln.variable[ var.VarName ] = {"Value" : var.X, "Id" : len(soln.variable) - 1}

            if extract_reduced_costs is True:
                for var in itertools.islice(pvars,self._last_native_var_idx+1):
                    soln.variable[ var.VarName ]["Rc"] = var.Rc

            if extract_duals is True:
                for con in cons:
                    # Pi attributes in Gurobi are the constraint duals
                    soln.constraint[ con.ConstrName ].dual = con.Pi                
                for con in qcons:
                    # QCPI attributes in Gurobi are the constraint duals
                    soln.constraint[ con.QCName ].dual = con.QCPi

            if extract_slacks is True:
                for con in cons:
                    soln.constraint[ con.ConstrName ].slack = con.Slack
                for con in qcons:
                    soln.constraint[ con.QCName ].slack = con.QCSlack
                # The above loops may include range constraints but will 
                # always report a slack of zero since gurobi transforms
                # range constraints by adding a slack variable in the following way
                # L <= f(x) <= U 
                # becomes
                # 0 <= U-f(x) <= U-L
                # becomes
                # U-f(x) == s
                # 0 <= s <= U-L
                # Therefore we need to check the value of the associated slack variable
                # with its upper bound to compute the original constraint slacks. To conform
                # with the other problem writers we return the slack value that is largest in 
                # magnitude (L-f(x) or U-f(x))
                for con,var_idx in self._range_con_var_pairs:
                    var = pvars[var_idx]
                    # U-f(x)
                    Us_ = var.X
                    # f(x)-L
                    Ls_ = var.UB-var.X
                    if Us_ > Ls_:
                        soln.constraint[ con.ConstrName ].slack = Us_
                    else:
                        soln.constraint[ con.ConstrName ].slack = -Ls_
                    


        results.solution.insert(soln)

        self.results = results

        # Done with the model object; free up some memory.
        self._last_native_var_idx = -1
        self._range_con_var_pairs = []
        del gprob, self._gurobi_instance

        # let the base class deal with returning results.
        return OptSolver._postsolve(self)


if not gurobi_python_api_exists:
    SolverFactory().deactivate('_gurobi_direct')
    SolverFactory().deactivate('_mock_gurobi_direct')



#
# Unit Tests for coopr.plugins.solvers.ASL
#
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
cooprdir = dirname(abspath(__file__))+os.sep+".."+os.sep+".."+os.sep
currdir = dirname(abspath(__file__))+os.sep

from nose.tools import nottest
import pyutilib.th as unittest
import pyutilib.services
import pyutilib.common
import coopr.plugins.solvers
import coopr.opt
import coopr
import xml
from coopr.opt import ResultsFormat, ProblemFormat, ConverterError

pyutilib.services.register_executable('cplexamp')
coopr.opt.SolverResults.default_print_options.ignore_time = True
try:
    cplex = coopr.plugins.solvers.CPLEX(keepfiles=True)
    cplex_available = (not cplex.executable() is None) and cplex.available(False)
    asl = coopr.plugins.solvers.ASL(keepfiles=True, options={'solver':'cplexamp'})
    cplexamp_available= cplex_available and (not asl.executable() is None) and asl.available(False)
except pyutilib.common.ApplicationError:
    cplexamp_available=False

try:
    pico_convert =  pyutilib.services.registered_executable("pico_convert")
    pico_convert_available= (not pico_convert is None)
except pyutilib.common.ApplicationError:
    pico_convert_available=False


def filter_cplex(line):
    return line.startswith("Message:")

class mock_all(unittest.TestCase):

    def setUp(self):
        self.do_setup(False)

    def do_setup(self,flag):
        global tmpdir
        tmpdir = os.getcwd()
        os.chdir(currdir)
        pyutilib.services.TempfileManager.sequential_files(0)
        pyutilib.services.TempfileManager.tempdir = currdir
        if flag:
            self.asl = coopr.plugins.solvers.ASL(keepfiles=True, options={'solver':'cplexamp'})
        else:
            self.asl = coopr.plugins.solvers.MockASL(keepfiles=True, options={'solver':'cplexamp'})
        self.asl.suffixes=['.*']

    def tearDown(self):
        global tmpdir
        pyutilib.services.TempfileManager.clear_tempfiles()
        os.chdir(tmpdir)
        pyutilib.services.TempfileManager.unique_files()

    def test_path(self):
        """ Verify that the ASL path is what is expected """
        if type(self.asl) is 'ASL':
            self.assertEqual(self.asl.executable.split(os.sep)[-1],"ASL"+coopr.util.executable_extension)

    def Xtest_solve1(self):
        """ Test ASL - test1.mps """
        results = self.asl.solve(currdir+"test1.mps", logfile=currdir+"test_solve1.log")
        results.write(filename=currdir+"test_solve1.txt", times=False, format='json')
        self.assertMatchesJsonBaseline(currdir+"test_solve1.txt", currdir+"test1_asl.txt")
        #os.remove(currdir+"test_solve1.log")

    def Xtest_solve2a(self):
        """ Test ASL - test1.mps """
        results = self.asl.solve(currdir+"test1.mps", rformat=ResultsFormat.soln, logfile=currdir+"test_solve2a.log")
        results.write(filename=currdir+"test_solve2a.txt", times=False, format='json')
        self.assertMatchesJsonBaseline(currdir+"test_solve2a.txt", currdir+"test1_asl.txt")
        #os.remove(currdir+"test_solve2a.log")

    def Xtest_solve2b(self):
        """ Test ASL - test1.mps """
        results = self.asl.solve(currdir+"test1.mps", pformat=ProblemFormat.mps, rformat=ResultsFormat.soln, logfile=currdir+"test_solve2b.log")
        results.write(filename=currdir+"test_solve2b.txt", times=False, format='json')
        self.assertMatchesJsonBaseline(currdir+"test_solve2b.txt", currdir+"test1_asl.txt")
        #os.remove(currdir+"test_solve2b.log")

    def Xtest_solve3(self):
        """ Test ASL - test2.lp """
        results = self.asl.solve(currdir+"test2.lp", logfile=currdir+"test_solve3.log", keepfiles=True)
        results.write(filename=currdir+"test_solve3.txt", times=False, format='json')
        self.assertMatchesJsonBaseline(currdir+"test_solve3.txt", currdir+"test2_asl.txt")
        if os.path.exists(currdir+"test2.solution.dat"):
            os.remove(currdir+"test2.solution.dat")
        #os.remove(currdir+"test_solve3.log")

    def test_solve4(self):
        """ Test ASL - test4.nl """
        results = self.asl.solve(currdir+"test4.nl", logfile=currdir+"test_solve4.log")
        results.write(filename=currdir+"test_solve4.txt", times=False, format='json')
        self.assertMatchesJsonBaseline(currdir+"test_solve4.txt", currdir+"test4_asl.txt", tolerance=1e-4)
        #os.remove(currdir+"test4.sol")
        #os.remove(currdir+"test_solve4.log")

    #
    # This test is disabled, but it's useful for interactively exercising
    # the option specifications of a solver
    #
    def Xtest_options(self):
        """ Test ASL options behavior """
        results = self.asl.solve(currdir+"bell3a.mps", logfile=currdir+"test_options.log", options="sec=0.1 foo=1 bar='a=b c=d' xx_zz=yy")
        results.write(filename=currdir+"test_options.txt",times=False)
        self.assertFileEqualsBaseline(currdir+"test_options.txt", currdir+  "test4_asl.txt")
        #os.remove(currdir+"test4.sol")
        #os.remove(currdir+"test_solve4.log")

    def Xtest_mock5(self):
        """ Mock Test ASL - test5.mps """
        if cplexamp_available:
            pass
        results = self.asl.solve(currdir+"test4.nl", logfile=currdir+"test_solve5.log", keepfiles=True)
        results.write(filename=currdir+"test_mock5.txt",times=False)
        self.assertFileEqualsBaseline(currdir+"test_mock5.txt", currdir+"test4_asl.txt")
        os.remove(currdir+"test4.sol")
        os.remove(currdir+"test_solve5.log")

    def test_error1(self):
        """ Bad results format """
        try:
            results = self.asl.solve(currdir+"test1.mps", format=ResultsFormat.sol)
            self.fail("test_error1")
        except ValueError:
            pass

    def test_error2(self):
        """ Bad solve option """
        try:
            results = self.asl.solve(currdir+"test1.mps", foo="bar")
            self.fail("test_error2")
        except ValueError:
            pass

    def test_error3(self):
        """ Bad solve option """
        try:
            results = self.asl.solve(currdir+"bad.mps", foo="bar")
            self.fail("test_error3")
        except ValueError:
            pass


@unittest.skipIf(not cplexamp_available, "The 'cplexamp' command is not available")
class mip_all(mock_all):

    def setUp(self):
        self.do_setup(True)


if __name__ == "__main__":
    unittest.main()



#
# Unit Tests for coopr.plugins.mip.CBC
#
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
cooprdir = dirname(abspath(__file__))+os.sep+".."+os.sep+".."+os.sep
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
import pyutilib.common
import pyutilib.services
from nose.tools import nottest
import coopr.plugins.mip
import coopr.opt
import coopr
import xml
from coopr.opt import ResultsFormat, ProblemFormat

coopr.opt.SolverResults.default_print_options.ignore_time = True
try:
    cbc = coopr.plugins.mip.CBC(keepfiles=True)
    cbc_available= (not cbc.executable() is None)
except pyutilib.common.ApplicationError:
    cbc_available=False
    pass

try:
    pico_convert =  pyutilib.services.registered_executable("pico_convert")
    pico_convert_available= (not pico_convert is None)
except pyutilib.common.ApplicationError:
    pico_convert_available=False



class mock_all(unittest.TestCase):

    def setUp(self):
        self.do_setup(False)

    def do_setup(self,flag):
        global tmpdir
        tmpdir = os.getcwd()
        os.chdir(currdir)
        pyutilib.services.TempfileManager.tempdir = currdir
        if flag:
            self.cbc = coopr.plugins.mip.CBC(keepfiles=True)
        else:
            self.cbc = coopr.plugins.mip.MockCBC(keepfiles=True)
        self.cbc.suffixes=['.*']

    def tearDown(self):
        global tmpdir
        pyutilib.services.TempfileManager.clear_tempfiles()
        os.chdir(tmpdir)

    def test_path(self):
        """ Verify that the CBC path is what is expected """
        if type(self.cbc) is 'CBC':
            self.failUnlessEqual(self.cbc.executable.split(os.sep)[-1],"CBC"+coopr.    util.executable_extension)

    def test_solve1(self):
        """ Test CBC - test1.mps """
        results = self.cbc.solve(currdir+"test1.mps",logfile=currdir+"test_solve1.log", solnfile=currdir+"test1.soln")
        results.write(filename=currdir+"test_solve1.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve1.txt", currdir+"test1_cbc.txt")
        os.remove(currdir+"test_solve1.log")
        os.remove(currdir+"test1.soln")

    def test_solve2a(self):
        """ Test CBC - test1.mps """
        results = self.cbc.solve(currdir+"test1.mps", rformat=ResultsFormat.soln, logfile=currdir+"test_solve2a.log", solnfile=currdir+"test1.soln")
        results.write(filename=currdir+"test_solve2a.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve2a.txt", currdir+"test1_cbc.txt")
        os.remove(currdir+"test_solve2a.log")
        os.remove(currdir+"test1.soln")

    def test_solve2b(self):
        """ Test CBC - test1.mps """
        results = self.cbc.solve(currdir+"test1.mps", pformat=ProblemFormat.mps, rformat=ResultsFormat.soln, logfile=currdir+"test_solve2b.log", solnfile=currdir+"test1.soln")
        results.write(filename=currdir+"test_solve2b.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve2b.txt", currdir+"test1_cbc.txt")
        os.remove(currdir+"test_solve2b.log")
        os.remove(currdir+"test1.soln")

    def test_solve3(self):
        """ Test CBC - test2.lp """
        results = self.cbc.solve(currdir+"test2.lp", logfile=currdir+"test_solve3.log", solnfile=currdir+"test2.soln")
        results.write(filename=currdir+"test_solve3.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve3.txt", currdir+"test2_cbc.txt")
        os.remove(currdir+"test_solve3.log")
        os.remove(currdir+"test2.soln")

    def test_solve4(self):
        """ Test CBC - test4.nl """
        try:
            results = self.cbc.solve(currdir+"test4.nl", rformat=ResultsFormat.sol, logfile=currdir+"test_solve4.log", solnfile=currdir+"test4.soln")
            results.write(filename=currdir+"test_solve4.txt",times=False)
            self.failUnlessFileEqualsBaseline(currdir+"test_solve4.txt", currdir+"test4_cbc.txt")
            #os.remove(currdir+"test_solve4.log")
            if os.path.exists(currdir+"ipconvert.lp"):
                os.remove(currdir+"ipconvert.lp")
        except coopr.opt.ConverterError:
            if pico_convert_available:
                raise "Expected conversion of test4.nl to a format that CBC can solve"

    #
    # This test is disabled, but it's useful for interactively exercising
    # the option specifications of a solver
    #
    def Xtest_options(self):
        """ Test CBC options behavior """
        results = self.cbc.solve(currdir+"bell3a.mps", logfile=currdir+"test_options.log", options="sec=0.1 foo=1 bar='a=b c=d' zz=yy")
        results.write(filename=currdir+"test_options.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_options.txt", currdir+  "test4_cbc.txt")
        os.remove(currdir+"test4.sol")
        os.remove(currdir+"test_solve4.log")

    def Xtest_prod(self):
        """ Test CBC - prod.mod """
        results = self.cbc.solve(currdir+"prod.mod", logfile=currdir+"test_prod.log", solnfile=currdir+"prod.soln")
        results.write(filename=currdir+"test_prod.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_prod.txt", currdir+"prod_cbc.txt")
        os.remove(currdir+"test_prod.log")
        if os.path.exists(currdir+"ipconvert.lp"):
            os.remove(currdir+"ipconvert.lp")
        os.remove(currdir+"prod.soln")

    def Xtest_error1(self):
        """ Bad results format """
        try:
            results = self.cbc.solve(currdir+"test1.mps", rformat=ResultsFormat.sol)
            self.fail("test_error1")
        except ValueError:
            pass

    def test_error2(self):
        """ Bad solve option """
        try:
            results = self.cbc.solve(currdir+"test1.mps", foo="bar")
            self.fail("test_error2")
        except ValueError:
            pass

    def test_error3(self):
        """ Bad solve option """
        try:
            results = self.cbc.solve(currdir+"bad.mps", foo="bar")
            self.fail("test_error3")
        except ValueError:
            pass


class mip_all(mock_all):

    def setUp(self):
        self.do_setup(True)

    def test_solve1a(self):
        """ Test CBC - test1.mps """
        results = self.cbc.solve(currdir+"test1.mps",logfile=currdir+"test_solve1a.log", solnfile=currdir+"test_solve1a.soln")
        results.write(filename=currdir+"test_solve1a.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve1a.txt", currdir+"test1_cbc.txt")
        os.remove(currdir+"test_solve1a.log")
        #os.remove(currdir+"test_solve1a.soln")

mip_all = unittest.skipIf(not cbc_available, "The 'cbc' command is not available")(mip_all)

if __name__ == "__main__":
    unittest.main()



#
# Unit Tests for util/misc
#
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
cooprdir = dirname(abspath(__file__))+"/../.."
currdir = dirname(abspath(__file__))+os.sep

import unittest
from nose.tools import nottest
#import coopr
import coopr.opt
import coopr.plugins.solvers
import pyutilib.services
import pyutilib.common
from pyutilib.component.core import alias


class TestSolver2(coopr.opt.OptSolver):

    alias('stest2')

    def __init__(self, **kwds):
        kwds['type'] = 'stest_type'
        coopr.opt.OptSolver.__init__(self,**kwds)

    def enabled(self):
        return False


class OptSolverDebug(unittest.TestCase):

    def setUp(self):
        pyutilib.services.TempfileManager.tempdir = currdir

    def tearDown(self):
        pyutilib.services.TempfileManager.clear_tempfiles()


    def test_solver_init1(self):
        """
        Verify the processing of 'type', 'name' and 'doc' options
        """
        ans = coopr.opt.SolverFactory("_mock_pico")
        self.assertEqual(type(ans), coopr.plugins.solvers.MockPICO)
        self.assertEqual(ans._doc, "pico OptSolver")

        ans = coopr.opt.SolverFactory("_mock_pico", doc="My Doc")
        self.assertEqual(type(ans), coopr.plugins.solvers.MockPICO)
        self.assertEqual(ans._doc, "My Doc")

        ans = coopr.opt.SolverFactory("_mock_pico", name="my name")
        self.assertEqual(type(ans), coopr.plugins.solvers.MockPICO)
        self.assertEqual(ans._doc, "my name OptSolver (type pico)")

    def test_solver_init2(self):
        """
        Verify that options can be passed in.
        """
        opt = {}
        opt['a'] = 1
        opt['b'] = "two"
        ans = coopr.opt.SolverFactory("_mock_pico", name="solver_init2", options=opt)
        self.assertEqual(ans.options['a'], opt['a'])
        self.assertEqual(ans.options['b'], opt['b'])

    def test_avail(self):
        ans = coopr.opt.SolverFactory("stest2")
        try:
            ans.available()
            self.fail("Expected exception for 'stest2' solver, which is disabled")
        except pyutilib.common.ApplicationError:
            pass



if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


__all__ = ['PyomoMIPConverter']

from pyutilib.component.core import *
from pyutilib.component.config import *
from pyutilib.component.executables import *
import pyutilib.services

from coopr.opt.base import *
from coopr.plugins.converter.pico import PicoMIPConverter


class PyomoMIPConverter(ManagedSingletonPlugin):

    implements(IProblemConverter)

    cmd = ExtensionPoint(IExternalExecutable)
    pico_converter = PicoMIPConverter()

    def __init__(self,**kwds):
        ManagedSingletonPlugin.__init__(self,**kwds)

    def can_convert(self, from_type, to_type):
        """Returns true if this object supports the specified conversion"""
        if from_type != ProblemFormat.pyomo:
            return False
        #
        # Return True for specific from/to pairs
        #
        if to_type in ( ProblemFormat.nl, ProblemFormat.cpxlp,
                        ProblemFormat.osil ):
            return True

        if to_type == ProblemFormat.mps and self.pico_converter.available():
            return True

        return False

    def apply(self, *args, **kwds):
        """
        Generate a NL or LP file from Pyomo, and then do subsequent
        conversions.
        """

        import coopr.pyomo

        capabilities = kwds.pop("capabilities", None)
        symbolic_solver_labels = kwds.pop('symbolic_solver_labels',False)

        if isinstance(args[2], basestring):
            instance = None
        else:
            instance = args[2]

        if args[1] is ProblemFormat.cpxlp:
            problem_filename = pyutilib.services.TempfileManager.create_tempfile(suffix = '.pyomo.lp')
            if instance is not None:
                (problem_filename, varmap) = instance.write(filename=problem_filename,
                                                            format=ProblemFormat.cpxlp,
                                                            solver_capability=capabilities,
                                                            symbolic_solver_labels=symbolic_solver_labels)
                return (problem_filename,),varmap # no map file is necessary
            else:
                ans = coopr.pyomo.scripting.pyomo2lp(['--save-model',problem_filename,args[2]])
                if ans.errorcode:
                    raise RuntimeError("pyomo2lp conversion returned nonzero error code (%s)" % ans.errorcode)
                model = ans.retval
                problem_filename = model.filename
                varmap = model.symbol_map
                return (problem_filename,),varmap

        elif args[1] in [ProblemFormat.mps, ProblemFormat.nl]:
            problem_filename = pyutilib.services.TempfileManager.create_tempfile(suffix = '.pyomo.nl')
            if not instance is None:
                (problem_filename, varmap) = instance.write(filename=problem_filename, 
                                                            format=ProblemFormat.nl, 
                                                            solver_capability=capabilities,
                                                            symbolic_solver_labels=symbolic_solver_labels)
            else:
                ans = coopr.pyomo.scripting.pyomo2nl(['--save-model',problem_filename,args[2]])
                if ans.errorcode:
                    raise RuntimeError("pyomo2lp conversion returned nonzero error code (%s)" % ans.errorcode)
                model = ans.retval
                problem_filename = model.filename
                varmap = model.symbol_map
            if args[1] is ProblemFormat.nl:
                return (problem_filename,),varmap
            #
            # Convert from NL to MPS
            #
            # TBD: We don't support a variable map file when going from NL to MPS within the PICO converter.
            # NOTE: this is a problem with the MPS writer that is provided by COIN-OR
            # NOTE: we should generalize this so it doesn't strictly depend on the PICO converter utility.
            #
            ans = self.pico_converter.apply(ProblemFormat.nl,ProblemFormat.mps,problem_filename)
            os.remove(problem_filename)
            return ans

        elif args[1] is ProblemFormat.osil:
            problem_filename = pyutilib.services.TempfileManager.create_tempfile(suffix='pyomo.osil')
            if instance:
                (problem_filename, varmap) = instance.write(filename=problem_filename,
                                                            format=ProblemFormat.osil,
                                                            solver_capability=capabilities,
                                                            symbolic_solver_labels=symbolic_solver_labels)
                return (problem_filename,), None
            else:
                msg = 'There is currently no script conversion available from\n'
                msg += 'Pyomo to OSiL format.'

                raise NotImplementedError(msg)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['PicoMIPConverter']

from coopr.opt.base import *

from pyutilib.component.core import *
from pyutilib.component.config import *
from pyutilib.component.executables import *
import pyutilib.services
import pyutilib.common
import pyutilib.subprocess


class PicoMIPConverter(ManagedSingletonPlugin):

    implements(IProblemConverter)

    executable = ExtensionPoint(IExternalExecutable)

    def __init__(self,**kwds):
        ManagedSingletonPlugin.__init__(self,**kwds)

    def can_convert(self, from_type, to_type):
        """Returns true if this object supports the specified conversion"""
        #
        # Test if the glpsol executable is available
        #
        if self.executable.service("pico_convert") is None:
            return False
        #
        # Return True for specific from/to pairs
        #
        if from_type == ProblemFormat.nl and to_type == ProblemFormat.cpxlp:
            return True
        if from_type == ProblemFormat.nl and to_type == ProblemFormat.mps:
            return True
        if from_type == ProblemFormat.mps and to_type == ProblemFormat.cpxlp:
            return True
        if from_type == ProblemFormat.cpxlp and to_type == ProblemFormat.mps:
            return True
        return False

    def available(self):
        cmd = self.executable.service("pico_convert")
        return not cmd is None

    def apply(self, *args, **kwargs):
        """
        Run the external pico_convert utility
        """
        if len(args) != 3:
            raise ConverterError("Cannot apply pico_convert with more than one filename or model")
        cmd = self.executable.service("pico_convert")
        if cmd is None:
            raise ConverterError("The 'pico_convert' application cannot be found")

        pico_convert_cmd = cmd.get_path()
        target=str(args[1])
        if target=="cpxlp":
            target="lp"
        # NOTE: if you have an extra "." in the suffix, the pico_convert program fails to output to the correct filename.
        output_filename = pyutilib.services.TempfileManager.create_tempfile(suffix = 'pico_convert.' + target)
        if not isinstance(args[2],basestring):
            fname= pyutilib.services.TempfileManager.create_tempfile(suffix= 'pico_convert.' +str(args[0]))
            args[2].write(filename=fname, format=args[1])
            cmd = pico_convert_cmd +" --output="+output_filename+" "+target+" "+fname
        else:
            cmd = pico_convert_cmd +" --output="+output_filename+" "+target
            for item in args[2:]:
                if not os.path.exists(item):
                    raise ConverterError("File "+item+" does not exist!")
                cmd = cmd + " "+item
        print("Running command: "+cmd)
        pyutilib.subprocess.run(cmd)
        if not os.path.exists(output_filename):       #pragma:nocover
            raise pyutilib.common.ApplicationError(\
                    "Problem launching 'pico_convert' to create "+output_filename)
        return (output_filename,),None # no variable map at the moment



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


import re
import os
import copy
import logging

from six import iteritems

import pyutilib.services
import pyutilib.common
import pyutilib.common
import pyutilib.component.core
import pyutilib.misc

from coopr.opt.base import *
from coopr.opt.base.solvers import _extract_version
from coopr.opt.results import *
from coopr.opt.solver import *
from coopr.plugins.solvers.mockmip import MockMIP


class PICO(OptSolver):
    """The PICO LP/MIP solver
    """

    pyutilib.component.core.alias('pico', doc='The PICO LP/MIP solver')

    def __new__(cls, *args, **kwds):
        try:
            mode = kwds['solver_io']
            if mode is None:
                mode = 'lp'
            del kwds['solver_io']
        except KeyError:
            mode = 'lp'
        #
        if mode  == 'lp':
            opt = SolverFactory('_pico_shell', **kwds)
            opt.set_problem_format(ProblemFormat.cpxlp)
            return opt
        #
        if mode == 'nl':
            # PICO does not accept all asl style command line options
            # (-s in particular, which is required for streaming output of 
            # all asl solvers). Therefore we need to send it through the cbc_shell
            # instead of ASL
            opt = SolverFactory('_pico_shell',**kwds)
            opt.set_problem_format(ProblemFormat.nl)
        elif mode == 'os':
            opt = SolverFactory('_ossolver', **kwds)
        else:
            logging.getLogger('coopr.plugins').error('Unknown IO type: %s' % mode)
            return
        opt.set_options('solver=PICO')
        return opt


class PICOSHELL(SystemCallSolver):
    """Shell interface to the PICO LP/MIP solver
    """

    pyutilib.component.core.alias('_pico_shell', doc='Shell interface to the PICO MIP solver')

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds["type"] = "pico"
        SystemCallSolver.__init__(self, **kwds)
        #
        # Setup valid problem formats, and valid results for each problem format
        #
        self._valid_problem_formats=[ProblemFormat.cpxlp, ProblemFormat.nl, ProblemFormat.mps]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.cpxlp] = [ResultsFormat.soln]
        self._valid_result_formats[ProblemFormat.nl] = [ResultsFormat.sol]
        self._valid_result_formats[ProblemFormat.mps] = [ResultsFormat.soln]
        self.set_problem_format(ProblemFormat.cpxlp)

        # Note: Undefined capabilities default to 'None'
        self._capabilities = pyutilib.misc.Options()
        self._capabilities.linear = True
        self._capabilities.integer = True
        #self._capabilities.sos1 = True
        #self._capabilities.sos2 = True

    def _default_results_format(self, prob_format):
        if prob_format == ProblemFormat.nl:
            return ResultsFormat.sol
        return ResultsFormat.soln

    def executable(self):
        executable = pyutilib.services.registered_executable("PICO")
        if executable is None:
            pyutilib.component.core.PluginGlobals.env().log.warning("Could not locate the 'PICO' executable, which is required for solver %s" % self.name)
            self.enable = False
            return None
        return executable.get_path()

    def version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        solver_exec = self.executable()
        if solver_exec is None:
            return _extract_version('')
        results = pyutilib.subprocess.run(
        [solver_exec, "--version"], timelimit=1)
        # 'PICO --version' seems to print 'pebble <version>, PICO <version>
        # we don't wan't the pebble version being advertised so we split
        # the string at the comma before extracting a version number. It
        # also exits with a nonzero return code so don't bother checking it.
        return _extract_version(results[1].split(',')[1])

    # Nothing needs to be done here
    #def _presolve(self, *args, **kwds):
    #    # let the base class handle any remaining keywords/actions.
    #    SystemCallSolver._presolve(self, *args, **kwds)

    def create_command_line(self,executable,problem_files):
        #
        # Define log file
        #
        if self.log_file is None:
            self.log_file = pyutilib.services.TempfileManager.create_tempfile(suffix="PICO.log")

        problem_filename_prefix = problem_files[0]
        if '.' in problem_filename_prefix:
            tmp = problem_filename_prefix.split('.')
            if len(tmp) > 2:
                problem_filename_prefix = '.'.join(tmp[:-1])
            else:
                problem_filename_prefix = tmp[0]
        if self._results_format is ResultsFormat.sol:
            self.soln_file = problem_filename_prefix+".sol"
        else:
            self.soln_file = problem_filename_prefix+".soln"

        #
        # Define results file
        #
        self.results_file = self.soln_file
        
        #
        # Eventually, these formats will be added to PICO...
        #
        #elif self._results_format == ResultsFormat.osrl:
            #self.results_file = self.tmpDir+os.sep+"PICO.osrl.xml"

        def _check_and_escape_options(options):
            for key, val in iteritems(self.options):
                tmp_k = str(key)
                _bad = ' ' in tmp_k

                tmp_v = str(val)
                if ' ' in tmp_v:
                    if '"' in tmp_v:
                        if "'" in tmp_v:
                            _bad = True
                        else:
                            tmp_v = "'" + tmp_v + "'"
                    else:
                        tmp_v = '"' + tmp_v + '"'

                if _bad:
                    raise ValueError("Unable to properly escape solver option:"
                                     "\n\t%s=%s" % (key, val) )
                yield (tmp_k, tmp_v)

        #
        # Define command line
        #
        cmd = [ executable ]
        if self._timer:
            cmd.insert(0, self._timer)

        if (self.options.mipgap is not None):
            raise ValueError("The mipgap parameter is currently not being "
                             "processed by PICO solver plugin")
        env=copy.copy(os.environ)
        if self._problem_format is None \
                or self._problem_format == ProblemFormat.nl:
            cmd.append(problem_files[0])
            cmd.append('-AMPL')

            opts = ['allowInfiniteIntegerVarBounds=true']
            if "debug" not in self.options:
                opts.append("debug=2")
            for key, val in _check_and_escape_options(self.options):
                if key == 'solver':
                    continue
                opts.append("%s=%s" % ( key, tmp ))
            opts.append('output='+self.soln_file)
            env["PICO_options"] = " ".join(opts)

        elif self._problem_format == ProblemFormat.cpxlp \
                or self._problem_format == ProblemFormat.mps:
            cmd.append('--allowInfiniteIntegerVarBounds=true')
            if "debug" not in self.options:
                cmd.extend(["--debug", "2"])
            for key, val in _check_and_escape_options(self.options):
                if key == 'mipgap':
                    continue
                cmd.extend(['--'+key, val])
            cmd.extend([ '--output', self.soln_file,
                         problem_files[0] ])

        return pyutilib.misc.Bunch(cmd=cmd, log_file=self.log_file, env=env)

    def process_logfile(self):
        """
        Process a logfile
        """
        results = SolverResults()

        #
        # Initial values
        #
        #results.solver.statistics.branch_and_bound.number_of_created_subproblems=0
        #results.solver.statistics.branch_and_bound.number_of_bounded_subproblems=0
        soln = Solution()
        soln.objective['__default_objective__'].value = None
        #
        # Process logfile
        #
        OUTPUT = open(self.log_file)
        output = "".join(OUTPUT.readlines())
        OUTPUT.close()
        #
        # Parse logfile lines
        #
        for line in output.split("\n"):
            tokens = re.split('[ \t]+',line.strip())
            if len(tokens) > 3 and tokens[0] == "ABORTED:":
                results.solver.status=SolverStatus.aborted
            elif len(tokens) > 1 and tokens[0].startswith("ERROR"):
                results.solver.status=SolverStatus.error
            elif len(tokens) == 3 and tokens[0] == 'Problem' and tokens[2].startswith('infeasible'):
                results.solver.termination_condition = TerminationCondition.infeasible
            elif len(tokens) == 2 and tokens[0] == 'Integer' and tokens[1] == 'Infeasible':
                results.solver.termination_condition = TerminationCondition.infeasible
            elif len(tokens) == 5 and tokens[0] == "Final" and tokens[1] == "Solution:":
                soln.objective['__default_objective__'].value = eval(tokens[4])
                soln.status = SolutionStatus.optimal
            elif len(tokens) == 3 and tokens[0] == "LP" and tokens[1] == "value=":
                soln.objective['__default_objective__'].value = eval(tokens[2])
                soln.status=SolutionStatus.optimal
                if results.problem.sense == ProblemSense.minimize:
                    results.problem.lower_bound = eval(tokens[2])
                else:
                    results.problem.upper_bound = eval(tokens[2])
            elif len(tokens) == 2 and tokens[0] == "Bound:":
                if results.problem.sense == ProblemSense.minimize:
                    results.problem.lower_bound = eval(tokens[1])
                else:
                    results.problem.upper_bound = eval(tokens[1])
            elif len(tokens) == 3 and tokens[0] == "Created":
                results.solver.statistics.branch_and_bound.number_of_created_subproblems = eval(tokens[1])
            elif len(tokens) == 3 and tokens[0] == "Bounded":
                results.solver.statistics.branch_and_bound.number_of_bounded_subproblems = eval(tokens[1])
            elif len(tokens) == 2 and tokens[0] == "sys":
                results.solver.system_time=eval(tokens[1])
            elif len(tokens) == 2 and tokens[0] == "user":
                results.solver.user_time=eval(tokens[1])
            elif len(tokens) == 3 and tokens[0] == "Solving" and tokens[1] == "problem:":
                results.problem.name = tokens[2]
            elif len(tokens) == 4 and tokens[2] == "constraints:":
                results.problem.number_of_constraints = eval(tokens[3])
            elif len(tokens) == 4 and tokens[2] == "variables:":
                results.problem.number_of_variables = eval(tokens[3])
            elif len(tokens) == 4 and tokens[2] == "nonzeros:":
                results.problem.number_of_nonzeros = eval(tokens[3])
            elif len(tokens) == 3 and tokens[1] == "Sense:":
                if tokens[2] == "minimization":
                    results.problem.sense = ProblemSense.minimize
                else:
                    results.problem.sense = ProblemSense.maximize

        if results.solver.status is SolverStatus.aborted:
            soln.optimality=SolutionStatus.unsure
        if soln.status is SolutionStatus.optimal:
            soln.gap=0.0
            results.problem.lower_bound = soln.objective['__default_objective__'].value
            results.problem.upper_bound = soln.objective['__default_objective__'].value

        if soln.status == SolutionStatus.optimal:
            results.solver.termination_condition = TerminationCondition.optimal

        if not results.solver.status is SolverStatus.error and \
            results.solver.termination_condition in [TerminationCondition.unknown,
                        #TerminationCondition.maxIterations,
                        #TerminationCondition.minFunctionValue,
                        #TerminationCondition.minStepLength,
                        TerminationCondition.globallyOptimal,
                        TerminationCondition.locallyOptimal,
                        TerminationCondition.optimal,
                        #TerminationCondition.maxEvaluations,
                        TerminationCondition.other]:
                results.solution.insert(soln)
        return results


    def process_soln_file(self,results):

        if self._results_format is ResultsFormat.sol:
            return

        # the only suffixes that we extract from PICO are
        # constraint duals. scan through the solver suffix
        # list and throw an exception if the user has
        # specified any others.
        extract_duals = False
        for suffix in self.suffixes:
            if re.match(suffix,"dual"):
                extract_duals = True
            else:
                raise RuntimeError("***PICO solver plugin cannot extract solution suffix="+suffix)

        #if os.path.exists(self.sol_file):
            #results_reader = ReaderFactory(ResultsFormat.sol)
            #results = results_reader(self.sol_file, results, results.solution(0))
            #return

        if not os.path.exists(self.soln_file):
            return
        if len(results.solution) == 0:
            return
        soln = results.solution(0)
        results.problem.num_objectives=1
        tmp=[]
        flag=False
        INPUT = open(self.soln_file,"r")
        lp_flag=None
        var_flag=True
        for line in INPUT:
            tokens = re.split('[ \t]+',line.strip())
            if len(tokens) == 0 or (len(tokens) == 1 and tokens[0]==''):
                continue
            if tokens[0] == "Objective":
                continue
            if lp_flag is None:
                lp_flag = (tokens[0] == "LP")
                continue
            if tokens[0] == "Dual" and tokens[1] == "solution:":
                var_flag=False
                # It looks like we've just been processing primal
                # variables.
                for (var,val) in tmp:
                    if var == 'ONE_VAR_CONSTANT':
                        continue
                    soln.variable[var] = {"Value" : val, "Id" : len(soln.variable)}
                tmp=[]
                continue
            if len(tokens) < 3:
                print("ERROR", line,tokens)
            tmp.append( (tokens[0],eval(tokens[2])) )
        if var_flag:
            for (var,val) in tmp:
                if var == 'ONE_VAR_CONSTANT':
                    continue
                soln.variable[var] = {"Value" : val, "Id" : len(soln.variable)}
        else:
            if (lp_flag is True) and (extract_duals is True):
                for (var,val) in tmp:
                    soln.constraint[var].dual = val
            else:
                for (var,val) in tmp:
                    soln.constraint[var] = val
        INPUT.close()


class MockPICO(PICOSHELL,MockMIP):
    """A Mock PICO solver used for testing
    """

    pyutilib.component.core.alias('_mock_pico')

    def __init__(self, **kwds):
        try:
            PICOSHELL.__init__(self,**kwds)
        except pyutilib.common.ApplicationError: #pragma:nocover
            pass                        #pragma:nocover
        MockMIP.__init__(self,"pico")

    def available(self, exception_flag=True):
        return PICOSHELL.available(self,exception_flag)

    def create_command_line(self,executable,problem_files):
        command = PICOSHELL.create_command_line(self,executable,problem_files)
        MockMIP.create_command_line(self,executable,problem_files)
        return command

    def executable(self):
        return MockMIP.executable(self)

    def _execute_command(self,cmd):
        return MockMIP._execute_command(self,cmd)



import os
import pickle
import sys
from nose.tools import timed

from pyutilib.common import ApplicationError
import pyutilib.th as unittest
from pyutilib.misc.pyyaml_util import *
from pyutilib.misc import Options

import coopr.pyomo.scripting.util as util
import coopr.plugins
from coopr.pyomo.scripting.util import apply_optimizer

currdir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(currdir)
problem_list = ['trivial_model']

def module_available(module):
    try:
        __import__(module)
        return True
    except ImportError:
        return False
def has_python(name):
    if module_available(name):
        return True
    return False

def has_gurobi_lp():
    try:
        gurobi = coopr.plugins.solvers.GUROBI(keepfiles=True)
        available = (not gurobi.executable() is None) and gurobi.available(False)
        return available
    except ApplicationError:
        raise
        return False
def has_cplex_lp():
    try:
        cplex = coopr.plugins.solvers.CPLEX(keepfiles=True)
        available = (not cplex.executable() is None) and cplex.available(False)
        return available
    except ApplicationError:
        return False
def has_glpk_lp():
    try:
        glpk = coopr.plugins.solvers.GLPK(keepfiles=True)
        available = (not glpk.executable() is None) and glpk.available(False)
        return available
    except ApplicationError:
        return False
def has_cbc_lp():
    try:
        cbc = coopr.plugins.solvers.CBC(keepfiles=True)
        available = (not cbc.executable() is None) and cbc.available(False)
        return available
    except ApplicationError:
        return False
def has_pico_lp():
    try:
        PICO = coopr.plugins.solvers.PICO(keepfiles=True)
        available = (not PICO.executable() is None) and PICO.available(False)
        return available
    except ApplicationError:
        return False

def has_cplex_nl():
    try:
        cplex = coopr.plugins.solvers.CPLEX(keepfiles=True)
        available = (not cplex.executable() is None) and cplex.available(False)
        asl = coopr.plugins.solvers.ASL(keepfiles=True, options={'solver':'cplexamp'})
        return available and (not asl.executable() is None) and asl.available(False)
    except ApplicationError:
        return False
def has_gurobi_nl():
    try:
        gurobi = coopr.plugins.solvers.GUROBI(keepfiles=True)
        available = (not gurobi.executable() is None) and gurobi.available(False)
        asl = coopr.plugins.solvers.ASL(keepfiles=True, options={'solver':'gurobi_ampl'})
        return available and (not asl.executable() is None) and asl.available(False)
    except ApplicationError:
        return False
def has_ipopt_nl():
    try:
        asl = coopr.plugins.solvers.ASL(keepfiles=True, options={'solver':'ipopt'})
        return (not asl.executable() is None) and asl.available(False)
    except ApplicationError:
        return False

writer_solver = []
if has_python('cplex'):
    writer_solver.append(('python','cplex',True))
else:
    writer_solver.append(('python','cplex',False))

if has_python('gurobipy'):
    writer_solver.append(('python','gurobi',True))
else:
    writer_solver.append(('python','gurobi',False))

if has_python('glpk'):
    writer_solver.append(('python','glpk',True))
else:
    writer_solver.append(('python','glpk',False))
    
if has_cplex_lp():
    writer_solver.append(('lp','cplex',True))
else:
    writer_solver.append(('lp','cplex',False))

if has_gurobi_lp():
    writer_solver.append(('lp','gurobi',True))
else:
    writer_solver.append(('lp','gurobi',False))
    
if has_glpk_lp():
    writer_solver.append(('lp','glpk',True))
else:
    writer_solver.append(('lp','glpk',False))

if has_cplex_nl():
    writer_solver.append(('nl','cplexamp',True))
elif has_gurobi_nl():
    writer_solver.append(('nl','gurobi_ampl',True))
elif has_ipopt_nl():
    writer_solver.append(('nl','ipopt',True))
else:
    writer_solver.append(('nl','ASL',False))

if has_cbc_lp():
    writer_solver.append(('lp','cbc',True))
else:
    writer_solver.append(('lp','cbc',False))
    
if has_pico_lp():
    writer_solver.append(('lp','pico',True))
else:
    writer_solver.append(('lp','pico',False))

def createTestMethod(pName,pickle_which,problem,solver,writer,do_test):

    @unittest.skipUnless(do_test,solver+"_"+writer+" not available")
    def testMethod(obj):
        options = Options()
        options.solver = solver
        options.solver_io = writer
        options.quiet = True
        data=Options(options=options)

        p = __import__(problem)
        model = p.define_model()
        inst = model.create()

        opt_data = apply_optimizer(data,instance=inst)

        f = open('junk.pickle','wb')
        if pickle_which == 'instance':
            inst.load(opt_data.results)
            pickle.dump(inst,f)
        elif pickle_which == 'results':
            pickle.dump(opt_data.results,f)
        f.close()

    return testMethod

def assignTests(cls):
    for PROBLEM in problem_list:
        for writer,solver,do_test in writer_solver:
            #for pickle_which in ['instance']:
            for pickle_which in ['instance','results']:
                attrName = "test_pickle_{0}_{1}_{2}_{3}".format(pickle_which,PROBLEM,solver,writer)
                setattr(cls,attrName,createTestMethod(attrName,pickle_which,PROBLEM,solver,writer,do_test))

@unittest.skipIf(writer_solver==[], "Can't find a solver.")
class PickleTest(unittest.TestCase):
    
    @classmethod
    def tearDownClass(self):
        try:
            os.unlink('junk.pickle')
        except Exception:
            pass

assignTests(PickleTest)

if __name__ == "__main__":
    unittest.main()




from coopr.pyomo import *

# Trivial model
def define_model():

    model = ConcreteModel()
    model.x = Var()
    model.obj = Objective(expr=model.x)
    model.con = Constraint(expr=model.x >= 1)

    return model



import pyutilib.th as unittest
from coopr.opt import *
from coopr.pyomo import *
import sys

try:
    import cplex
    cplexpy_available = True
except ImportError:
    cplexpy_available = False

diff_tol = 1e-4

class CPLEXDirectTests(unittest.TestCase):

    def setUp(self):
        self.stderr = sys.stderr
        sys.stderr = None

    def tearDown(self):
        sys.stderr = self.stderr

    @unittest.skipIf(not cplexpy_available,"The 'cplex' python bindings are not available")
    def test_infeasible_lp(self):
        self.opt = SolverFactory("cplex",solver_io="python")
        
        model = AbstractModel()
        model.X = Var(within=NonNegativeReals)
        model.C1 = Constraint(expr= model.X==1)
        model.C2 = Constraint(expr= model.X==2)
        model.O = Objective(expr= model.X)

        instance = model.create()
        results = self.opt.solve(instance)

        self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasible)

    @unittest.skipIf(not cplexpy_available,"The 'cplex' python bindings are not available")
    def test_unbounded_lp(self):
        self.opt = SolverFactory("cplex",solver_io="python")
        
        model = AbstractModel()
        model.X = Var()
        model.O = Objective(expr= model.X)

        instance = model.create()
        results = self.opt.solve(instance)

        self.assertEqual(results.solver.termination_condition, TerminationCondition.unbounded)

    @unittest.skipIf(not cplexpy_available,"The 'cplex' python bindings are not available")
    def test_optimal_lp(self):
        self.opt = SolverFactory("cplex",solver_io="python")
        
        model = AbstractModel()
        model.X = Var(within=NonNegativeReals)
        model.O = Objective(expr= model.X)

        instance = model.create()
        results = self.opt.solve(instance)

        self.assertEqual(results.solution.status, SolutionStatus.optimal)

    @unittest.skipIf(not cplexpy_available,"The 'cplex' python bindings are not available")
    def test_get_duals_lp(self):
        self.opt = SolverFactory("cplex",solver_io="python")
        
        model = AbstractModel()
        model.X = Var(within=NonNegativeReals)
        model.Y = Var(within=NonNegativeReals)

        model.C1 = Constraint(expr= 2*model.X + model.Y >= 8 )
        model.C2 = Constraint(expr= model.X + 3*model.Y >= 6 )

        model.O = Objective(expr= model.X + model.Y)

        instance = model.create()
        instance.load(self.opt.solve(instance,suffixes=['dual']))

        self.assertAlmostEqual(instance.dual.getValue(instance.C1), 0.4)
        self.assertAlmostEqual(instance.dual.getValue(instance.C2), 0.2)




    @unittest.skipIf(not cplexpy_available,"The 'cplex' python bindings are not available")
    def test_infeasible_mip(self):
        self.opt = SolverFactory("cplex",solver_io="python")
        
        model = AbstractModel()
        model.X = Var(within=NonNegativeIntegers)
        model.C1 = Constraint(expr= model.X==1)
        model.C2 = Constraint(expr= model.X==2)
        model.O = Objective(expr= model.X)

        instance = model.create()
        results = self.opt.solve(instance)

        self.assertEqual(results.solver.termination_condition, TerminationCondition.infeasible)

    @unittest.skipIf(not cplexpy_available,"The 'cplex' python bindings are not available")
    def test_unbounded_mip(self):
        self.opt = SolverFactory("cplex",solver_io="python")
        
        model = AbstractModel()
        model.X = Var(within=Integers)
        model.O = Objective(expr= model.X)

        instance = model.create()
        results = self.opt.solve(instance)

        self.assertEqual(results.solver.termination_condition, TerminationCondition.unbounded)

    @unittest.skipIf(not cplexpy_available,"The 'cplex' python bindings are not available")
    def test_optimal_mip(self):
        self.opt = SolverFactory("cplex",solver_io="python")
        
        model = AbstractModel()
        model.X = Var(within=NonNegativeIntegers)
        model.O = Objective(expr= model.X)

        instance = model.create()
        results = self.opt.solve(instance)

        self.assertEqual(results.solution.status, SolutionStatus.optimal)





if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


import re
import os
import copy

import pyutilib.services
import pyutilib.common
import pyutilib.common
import pyutilib.component.core
import pyutilib.component.config
import pyutilib.misc

from coopr.opt.base import *
from coopr.opt.base.solvers import _extract_version
from coopr.opt.results import *
from coopr.opt.solver import *
from coopr.plugins.solvers.mockmip import MockMIP

try:
    unicode
except:
    basestring = str


class ASL(SystemCallSolver):
    """A generic optimizer that uses the AMPL Solver Library to interface with applications.
    """

    pyutilib.component.core.alias('_asl', doc='Shell interface to a generic optimizer that uses the AMPL Solver Library to interface with application')

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds["type"] = "asl"
        SystemCallSolver.__init__(self, **kwds)
        #
        # Setup valid problem formats, and valid results for each problem format.
        # Also set the default problem and results formats.
        #
        self._valid_problem_formats=[ProblemFormat.nl]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.nl] = [ResultsFormat.sol]
        self.set_problem_format(ProblemFormat.nl)
        
        # Note: Undefined capabilities default to 'None'
        self._capabilities = pyutilib.misc.Options()
        self._capabilities.linear = True
        self._capabilities.integer = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def _default_results_format(self, prob_format):
        return ResultsFormat.sol

    def executable(self):
        #
        # We register the ASL executables dynamically, since _any_ ASL solver could be
        # executed by this solver.
        #
        try:
            if self.options.solver is None:
                raise pyutilib.component.config.OptionError('ERROR')
            pyutilib.services.register_executable(self.options.solver)
        except pyutilib.component.config.OptionError:
            raise ValueError("No solver option specified for ASL solver interface")
        executable = pyutilib.services.registered_executable(self.options.solver)
        if executable is None:
            pyutilib.component.core.PluginGlobals.env().log.warning("Could not locate the '%s' executable, which is required for solver %s" % (self.options.solver, self.name))
            self.enable = False
            return None
        return executable.get_path()

    def version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        solver_exec = self.executable()
        if solver_exec is None:
            return _extract_version('')
        results = pyutilib.subprocess.run( [solver_exec,"-v"], timelimit=1 )
        return _extract_version(results[1])

    def create_command_line(self, executable, problem_files):
 
        assert(self._problem_format == ProblemFormat.nl)
        assert(self._results_format == ResultsFormat.sol)

        #
        # Define log file
        #
        if self.log_file is None:
            self.log_file = pyutilib.services.TempfileManager.create_tempfile(suffix="_asl.log")
        fname = problem_files[0]
        if '.' in fname:
            tmp = fname.split('.')
            if len(tmp) > 2:
                fname = '.'.join(tmp[:-1])
            else:
                fname = tmp[0]
        self.soln_file = fname+".sol"

        #
        # Define results file
        #
        self.results_file = self.soln_file
        
        #
        # Define command line
        #
        env=copy.copy(os.environ)

        #
        # Merge the COOPR_AMPLFUNC (externals defined within
        # Coopr/Pyomo) with any user-specified external function
        # libraries
        #
        if 'COOPR_AMPLFUNC' in env:
            if 'AMPLFUNC' in env:
                env['AMPLFUNC'] += "\n" + env['COOPR_AMPLFUNC']
            else:
                env['AMPLFUNC'] = env['COOPR_AMPLFUNC']

        cmd = [executable, '-s', problem_files[0]]
        if self._timer:
            cmd.insert(0, self._timer)
        
        # GAH: I am going to re-add the code by Zev that passed options through
        # to the command line. Setting the environment variable in this way does
        # NOT work for solvers like cplex and gurobi because the are looking for 
        # an environment variable called cplex_options / gurobi_options. However
        # the options.solver name for these solvers is cplexamp / gurobi_ampl 
        # (which creates a cplexamp_options and gurobi_ampl_options env variable).
        # Because of this, I think the only reliable way to pass options for any 
        # solver is by using the command line
        opt=[]
        for key in self.options:
            if key is 'solver':
                continue
            if isinstance(self.options[key],basestring) and ' ' in self.options[key]:
                opt.append(key+"=\""+str(self.options[key])+"\"")
                cmd.append(str(key)+"="+str(self.options[key]))
            elif key == 'subsolver':
                opt.append("solver="+str(self.options[key]))
                cmd.append(str(key)+"="+str(self.options[key]))
            else:
                opt.append(key+"="+str(self.options[key]))
                cmd.append(str(key)+"="+str(self.options[key]))

        envstr = "%s_options" % self.options.solver
        # Merge with any options coming in through the environment
        env[envstr] = " ".join(opt)
            
        return pyutilib.misc.Bunch(cmd=cmd, log_file=self.log_file, env=env)

    def Xprocess_soln_file(self,results):
        """
        Process the SOL file
        """
        if os.path.exists(self.soln_file):
            results_reader = ReaderFactory(ResultsFormat.sol)
            results = results_reader(self.soln_file, results, results.solution(0))
            return


class MockASL(ASL,MockMIP):
    """A Mock ASL solver used for testing
    """

    pyutilib.component.core.alias('_mock_asl')

    def __init__(self, **kwds):
        try:
            ASL.__init__(self,**kwds)
        except pyutilib.common.ApplicationError: #pragma:nocover
            pass                        #pragma:nocover
        MockMIP.__init__(self,"asl")

    def available(self, exception_flag=True):
        return ASL.available(self,exception_flag)

    def create_command_line(self,executable,problem_files):
        command = ASL.create_command_line(self,executable,problem_files)
        MockMIP.create_command_line(self,executable,problem_files)
        return command

    def executable(self):
        return MockMIP.executable(self)

    def _execute_command(self,cmd):
        return MockMIP._execute_command(self,cmd)



#
# Unit Tests for coopr.plugins.mip.GLPK
#
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
cooprdir = dirname(abspath(__file__))+os.sep+".."+os.sep+".."+os.sep
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
import pyutilib.common
import pyutilib.services
from nose.tools import nottest
import coopr.plugins.mip
import coopr.opt
import coopr
import xml
from coopr.opt import ResultsFormat, ProblemFormat

coopr.opt.SolverResults.default_print_options.ignore_time = True
try:
    glpk = coopr.plugins.mip.GLPK(keepfiles=True)
    glpk_available= (not glpk.executable() is None)
except IOError:
    glpk_available=False
    pass
except pyutilib.common.ApplicationError:
    glpk_available=False
    pass

try:
    pico_convert =  pyutilib.services.registered_executable("pico_convert")
    pico_convert_available= (not pico_convert is None)
except pyutilib.common.ApplicationError:
    pico_convert_available=False



class mock_all(unittest.TestCase):

    def setUp(self):
        self.do_setup(False)
        pyutilib.services.TempfileManager.tempdir = currdir

    def do_setup(self,flag):
        if flag:
            self.glpk = coopr.plugins.mip.GLPK(keepfiles=True)
        else:
            self.glpk = coopr.plugins.mip.MockGLPK(keepfiles=True)
        self.glpk.suffixes=['.*']

    def tearDown(self):
        if os.path.exists(currdir+"glpk.soln"):
            os.remove(currdir+"glpk.soln")
        pyutilib.services.TempfileManager.clear_tempfiles()

    def test_path(self):
        """ Verify that the GLPK path is what is expected """
        if type(self.glpk) is 'GLPK':
            self.failUnlessEqual(self.glpk.executable.split(os.sep)[-1],"GLPK"+coopr.util.executable_extension)

    def test_solve1(self):
        """ Test GLPK - test1.mps """
        results = self.glpk.solve(currdir+"test1.mps",logfile=currdir+"test_solve1.log", solnfile=currdir+"test1.soln")
        results.write(filename=currdir+"test_solve1.txt", times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve1.txt", currdir+"test1_glpk.txt")
        os.remove(currdir+"test_solve1.log")
        os.remove(currdir+"test1.soln")

    def test_solve2a(self):
        """ Test GLPK - test1.mps """
        results = self.glpk.solve(currdir+"test1.mps", rformat=ResultsFormat.soln, logfile=currdir+"test_solve2a.log", solnfile=currdir+"test1.soln")
        results.write(filename=currdir+"test_solve2a.txt", times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve2a.txt", currdir+"test1_glpk.txt")
        os.remove(currdir+"test_solve2a.log")
        os.remove(currdir+"test1.soln")

    def test_solve2b(self):
        """ Test GLPK - test1.mps """
        results = self.glpk.solve(currdir+"test1.mps", pformat=ProblemFormat.mps, rformat=ResultsFormat.soln, logfile=currdir+"test_solve2b.log", solnfile=currdir+"test1.soln")
        results.write(filename=currdir+"test_solve2b.txt", times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve2b.txt", currdir+"test1_glpk.txt")
        os.remove(currdir+"test_solve2b.log")
        os.remove(currdir+"test1.soln")

    def test_solve3(self):
        """ Test GLPK - test2.lp """
        results = self.glpk.solve(currdir+"test2.lp", logfile=currdir+"test_solve3.log", solnfile=currdir+"test2.soln")
        results.write(filename=currdir+"test_solve3.txt", times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve3.txt", currdir+"test2_glpk.txt")
        os.remove(currdir+"test_solve3.log")
        os.remove(currdir+"test2.soln")

    def test_solve4(self):
        """ Test GLPK - test4.nl """
        if pico_convert_available:
            results = self.glpk.solve(currdir+"test4.nl", rformat=ResultsFormat.soln, logfile=currdir+"test_solve4.log")
            results.write(filename=currdir+"test_solve4.txt",times=False)
            self.failUnlessFileEqualsBaseline(currdir+"test_solve4.txt", currdir+"test4_glpk.txt")
            #os.remove(currdir+"test4.soln")
            os.remove(currdir+"test_solve4.log")
        else:
            try:
                results = self.glpk.solve(currdir+"test4.nl", rformat=ResultsFormat.soln, logfile=currdir+"test_solve4.log")
            except coopr.opt.ConverterError:
                pass


    def Xtest_solve5(self):
        """ Test GLPK - test5.mps """
        results = self.glpk.solve(currdir+"test5.mps", rformat=ResultsFormat.soln, logfile=currdir+"test_solve5.log", solnfile=currdir+"test5.soln",timelimit=300)
        self.failUnlessEqual(results.solution(0).status, coopr.opt.SolutionStatus.bestSoFar)
        os.remove(currdir+"test_solve5.log")
        os.remove(currdir+"test5.soln")

    def test_diet(self):
        """ Test GLPK - diet.mod """
        results = self.glpk.solve(currdir+"diet.mod", logfile=currdir+"test_diet.log", solnfile=currdir+"diet.soln")
        results.write(filename=currdir+"test_diet.txt", times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_diet.txt", currdir+"diet_glpk.txt")
        os.remove(currdir+"test_diet.log")
        if os.path.exists(currdir+"ipconvert.lp"):
            os.remove(currdir+"ipconvert.lp")
        os.remove(currdir+"diet.soln")

    #
    # This test is disabled, but it's useful for interactively exercising
    # the option specifications of a solver
    #
    def Xtest_options(self):
        """ Test GLPK options behavior """
        results = self.glpk.solve(currdir+"bell3a.mps", logfile=currdir+"test_options.log", options="maxCPUMinutes=0.1 foo=1 bar='a=b c=d' zz=yy")
        results.write(filename=currdir+"test_options.txt",times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_options.txt", currdir+  "test4_glpk.txt")
        #os.remove(currdir+"test4.sol")
        #os.remove(currdir+"test_solve4.log")

    def Xtest_error1(self):
        """ Bad results format """
        try:
            results = self.glpk.solve(currdir+"test1.mps", rformat=ResultsFormat.sol)
            self.fail("test_error1")
        except ValueError:
            pass

    def test_error2(self):
        """ Bad solve option """
        try:
            results = self.glpk.solve(currdir+"test1.mps", foo="bar")
            self.fail("test_error2")
        except ValueError:
            pass

    def test_error3(self):
        """ Bad solve option """
        try:
            results = self.glpk.solve(currdir+"bad.mps", foo="bar")
            self.fail("test_error3")
        except ValueError:
            pass


class mip_all(mock_all):

    def setUp(self):
        self.do_setup(True)

    def test_solve1a(self):
        """ Test GLPK - test1.mps """
        results = self.glpk.solve(currdir+"test1.mps")
        results.write(filename=currdir+"test_solve1.txt", times=False)
        self.failUnlessFileEqualsBaseline(currdir+"test_solve1.txt", currdir+"test1_glpk.txt")
        if os.path.exists(currdir+"glpk.log"):
            os.remove(currdir+"glpk.log")
        if os.path.exists(currdir+"glpk.soln"):
            os.remove(currdir+"glpk.soln")

mip_all = unittest.skipIf(not glpk_available, "The 'glpsol' command is not available")(mip_all)

if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import os
import os.path

import pyutilib.autotest
from pyutilib.component.core import *
import pyutilib.services
from pyutilib.misc import Options

import coopr.opt


class CooprMIPTestDriver(Plugin):

    implements(pyutilib.autotest.ITestDriver)
    alias('coopr.mip')

    def setUpClass(self, cls, options):
        try:
            cls.pico_convert =  pyutilib.services.registered_executable("pico_convert")
            cls.pico_convert_available= (not cls.pico_convert is None)
        except pyutilib.common.ApplicationError:
            cls.pico_convert_available=False

    def tearDownClass(self, cls, options):
        pass

    def setUp(self, testcase, options):
        global tmpdir
        tmpdir = os.getcwd()
        os.chdir(options.currdir)
        pyutilib.services.TempfileManager.push()
        pyutilib.services.TempfileManager.sequential_files(0)
        pyutilib.services.TempfileManager.tempdir = options.currdir
        #
        if ':' in options.solver:
            solver, sub_solver = options.solver.split(':')
            if options.solver_options is None:
                _options = Options()
            else:
                _options = options.solver_options
            _options.solver = sub_solver
            testcase.opt = coopr.opt.SolverFactory(solver, options=_options)
        else:
            testcase.opt = coopr.opt.SolverFactory(options.solver, options=options.solver_options)
        if testcase.opt is None or not testcase.opt.available(False):
            testcase.skipTest('Solver %s is not available' % options.solver)
        else:
            testcase.opt.suffixes = ['.*']

    def tearDown(self, testcase, options):
        global tmpdir
        pyutilib.services.TempfileManager.pop()
        os.chdir(tmpdir)
        pyutilib.services.TempfileManager.unique_files()

    def run_test(self, testcase, name, options):
        # TODO dirty hack for broken GLPK MIP solve - more elegant solution?
        if 'glpk_test1' in name:
            import subprocess
            try:
                p = subprocess.Popen(['glpsol', '--version'], stdout=subprocess.PIPE)
                stdoutdata, stderrdata = p.communicate()
                vstr = stdoutdata.split("\n")[0].split(" ")[-1]
                if vstr == '4.40':
                    testcase.skipTest("Bug in GLPK 4.40 for MIP solving on this test case")
            except OSError:
                pass # Couldn't find glpsol; ignore, since we're (probably) using mock glpk
        if options.verbose or options.debug:
            print("Test %s - Running coopr.opt solver with %s" % (name, str(options)))
        if not options.use_pico_convert or (options.use_pico_convert and testcase.pico_convert_available):
            try:
                if options.results_format:
                    if options.verbose or options.debug:
                        print("Running with results format %s" % options.results_format)
                    results = testcase.opt.solve(options.currdir+options.files, rformat=coopr.opt.ResultsFormat(options.results_format), logfile=options.currdir+name+".log")
                else:
                    if options.verbose or options.debug:
                        print("Running with default results format")
                    results = testcase.opt.solve(options.currdir+options.files, logfile=options.currdir+name+".log")
                    if options.verbose or options.debug:
                        print("-----------------------------------------------------------------")
                        print("Results: %s" % results)
                        print("-----------------------------------------------------------------")
            except coopr.opt.ConverterError:
                testcase.skipTest('Cannot convert problem files: %s' % options.files)
            baseline = pyutilib.misc.load_json( pyutilib.misc.extract_subtext( options.baseline ) )
            if options.tolerance is None:
                tol = 1e-7
            else:
                tol = options.tolerance
            pyutilib.misc.compare_json_repn( baseline, results.json_repn(), tolerance=tol, exact=False)
        else:
            try:
                if options.results_format:
                    results = testcase.opt.solve(options.currdir+options.files, rformat=coopr.opt.ResultsFormat(options.results_format), logfile=options.currdir+name+".log")
                else:
                    results = testcase.opt.solve(options.currdir+options.files, logfile=options.currdir+name+".log")
            except coopr.opt.ConverterError:
                testcase.skipTest('Cannot convert problem files: %s' % options.files)
        if (os.path.exists(options.currdir+name+".log") is True) and (not options.debug):
            os.remove(options.currdir+name+".log")
        else:
            print("Solver log file saved in %s" % options.currdir+name+".log")



#
# Unit Tests for coopr.opt.base.convert
#
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
cooprdir = dirname(abspath(__file__))+os.sep+".."+os.sep+".."+os.sep
currdir = dirname(abspath(__file__))+os.sep

import re
from nose.tools import nottest
import coopr.plugins
from coopr.opt import ProblemFormat, ConverterError
import coopr
import pyutilib.th as unittest
import pyutilib.services
import pyutilib.common
import pyutilib.component.core
import pyutilib.component.config
import xml
import filecmp
import shutil


def filter(line):
    return 'Problem' in line or line.startswith('NAME')

class MockArg(object):

    def __init__(self):
        pass

    def valid_problem_types(self):
        return [ProblemFormat.pyomo]

    def write(self,filename="", format=None, solver_capability=None, symbolic_solver_labels=False):
        return (filename,None)

class MockArg2(MockArg):

    def valid_problem_types(self):
        return [ProblemFormat.nl]

    def write(self,filename="", format=None, solver_capability=None, symbolic_solver_labels=False):
        OUTPUT=open(filename,"w")
        INPUT=open(currdir+"test4.nl")
        for line in INPUT:
            OUTPUT.write(line)
        OUTPUT.close()
        INPUT.close()
        return (filename,None)

class MockArg3(MockArg):

    def valid_problem_types(self):
        return [ProblemFormat.mod]

    def write(self,filename="", format=None, solver_capability=None, symbolic_solver_labels=False):
        return (filename,None)

class MockArg4(MockArg):

    def write(self,filename="", format=None, solver_capability=None, symbolic_solver_labels=False):
        OUTPUT=open(filename,"w")
        INPUT=open(currdir+"test4.nl")
        for line in INPUT:
            OUTPUT.write(line)
        OUTPUT.close()
        INPUT.close()
        return (filename,None)


class Test(unittest.TestCase):

    def setUp(self):
        pyutilib.services.TempfileManager.tempdir = currdir

    def tearDown(self):
        pyutilib.services.TempfileManager.clear_tempfiles()
        #
        # Reset all options
        #
        for ep in pyutilib.component.core.ExtensionPoint(pyutilib.component.config.IOption):
            ep.reset()

    def test_nl_nl1(self):
        """ Convert from NL to NL """
        ans = coopr.opt.convert_problem( ("test4.nl",), None, [ProblemFormat.nl])
        self.assertEqual(ans[0],("test4.nl",))

    def test_nl_nl2(self):
        """ Convert from NL to NL """
        ans = coopr.opt.convert_problem( ("test4.nl","tmp.nl"), None, [ProblemFormat.nl])
        self.assertEqual(ans[0],("test4.nl","tmp.nl"))

    def test_nl_lp1(self):
        """ Convert from NL to LP """
        try:
            ans = coopr.opt.convert_problem( (currdir+"test4.nl",), None, [ProblemFormat.cpxlp])
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("pico_convert") is None:
                self.fail("Unexpected ApplicationError - pico_convert is enabled but not available: '%s'" % str(err))
            return
        except ConverterError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("pico_convert") is None:
                self.fail("Unexpected ConverterError - pico_convert is enabled but not available: '%s'" % str(err))
            return
        self.assertEqual(ans[0][0][-15:],"pico_convert.lp")
        self.assertFileEqualsBaseline(ans[0][0], currdir+"test1_convert.lp")

    def test_mod_lp1(self):
        """ Convert from MOD to LP """
        try:
            ans = coopr.opt.convert_problem( (currdir+"test3.mod",), None, [ProblemFormat.cpxlp])
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("glpsol") is None:
                self.fail("Unexpected ApplicationError - glpsol is enabled but not available: '%s'" % str(err))
            return
        except ConverterError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("glpsol") is None:
                self.fail("Unexpected ConverterError - glpsol is enabled but not available: '%s'" % str(err))
            return
        self.assertTrue(ans[0][0].endswith("glpsol.lp"))
        self.assertFileEqualsBaseline(ans[0][0], currdir+"test2_convert.lp", filter=filter)

    def test_mod_lp2(self):
        """ Convert from MOD+DAT to LP """
        try:
            ans = coopr.opt.convert_problem( (currdir+"test5.mod",currdir+"test5.dat"), None, [ProblemFormat.cpxlp])
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("glpsol") is None:
                self.fail("Unexpected ApplicationError - glpsol is enabled but not available: '%s'" % str(err))
            return
        except ConverterError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("glpsol") is None:
                self.fail("Unexpected ConverterError - glpsol is enabled but not available: '%s'" % str(err))
            return
        self.assertTrue(ans[0][0].endswith("glpsol.lp"))
        self.assertFileEqualsBaseline(ans[0][0], currdir+"test3_convert.lp", filter=filter)

    def test_mod_nl1(self):
        """ Convert from MOD to NL """
        try:
            ans = coopr.opt.convert_problem( (currdir+"test3.mod",), None, [ProblemFormat.nl])
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("ampl") is None:
                self.fail("Unexpected ApplicationError - ampl is enabled but not available: '%s'" % str(err))
            return
        except ConverterError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("ampl") is None:
                self.fail("Unexpected ConverterError - ampl is enabled but not available: '%s'" % str(err))
            return
        self.assertTrue(ans[0][0].endswith('.nl'))
        #self.assertFileEqualsBinaryFile(ans[0][0], currdir+"test_mod_nl1.nl")

    def test_mod_nl2(self):
        """ Convert from MOD+DAT to NL """
        try:
            ans = coopr.opt.convert_problem( (currdir+"test5.mod",currdir+"test5.dat"), None, [ProblemFormat.nl])
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("ampl") is None:
                self.fail("Unexpected ApplicationError - ampl is enabled but not available: '%s'" % str(err))
            return
        except ConverterError:
            err = sys.exc_info()[1]
            if not pyutilib.services.registered_executable("ampl") is None:
                self.fail("Unexpected ConverterError - ampl is enabled but not available: '%s'" % str(err))
            return
        self.assertTrue(ans[0][0].endswith('.nl'))
        #self.assertFileEqualsBaseline(ans[0][0], currdir+"test_mod_nl2.nl")

    def test_mock_lp1(self):
        """ Convert from Pyomo to LP """
        arg=MockArg()
        ans = coopr.opt.convert_problem( (arg,ProblemFormat.cpxlp,arg), None, [ProblemFormat.cpxlp])
        self.assertNotEqual(re.match(".*tmp.*pyomo.lp$",ans[0][0]), None)

    def test_pyomo_lp1(self):
        """ Convert from Pyomo to LP with file"""
        ans = coopr.opt.convert_problem( (currdir+'model.py',ProblemFormat.cpxlp,), None, [ProblemFormat.cpxlp])
        self.assertNotEqual(re.match(".*tmp.*pyomo.lp$",ans[0][0]), None)

    def test_mock_lp2(self):
        """ Convert from NL to LP """
        arg=MockArg2()
        try:
            ans = coopr.opt.convert_problem( (arg,), None, [ProblemFormat.cpxlp])
        except ConverterError:
            err = sys.exc_info()[1]
            if pyutilib.services.registered_executable("pico_convert") is None:
                return
            else:
                self.fail("Expected ApplicationError because pico_convert is not available: '%s'" % str(err))
        self.assertEqual(ans[0][0][-15:],"pico_convert.lp")
        os.remove(ans[0][0])

    def test_mock_mps1(self):
        """ Convert from Pyomo to MPS """
        arg=MockArg4()
        try:
            ans = coopr.opt.convert_problem( (arg,ProblemFormat.mps,arg), None, [ProblemFormat.mps])
        except ConverterError:
            err = sys.exc_info()[1]
            if pyutilib.services.registered_executable("pico_convert") is None:
                return
            else:
                self.fail("Expected ApplicationError because pico_convert is not available: '%s'" % str(err))
        self.assertEqual(ans[0][0][-16:],"pico_convert.mps")
        os.remove(ans[0][0])

    def test_pyomo_mps1(self):
        """ Convert from Pyomo to MPS with file"""
        try:
            ans = coopr.opt.convert_problem( (currdir+'model.py',ProblemFormat.mps,), None, [ProblemFormat.mps])
        except ConverterError:
            err = sys.exc_info()[1]
            if pyutilib.services.registered_executable("pico_convert") is None:
                return
            else:
                self.fail("Expected ApplicationError because pico_convert is not available: '%s'" % str(err))
        self.assertEqual(ans[0][0][-16:],"pico_convert.mps")
        os.remove(ans[0][0])

    def test_mock_nl1(self):
        """ Convert from Pyomo to NL """
        arg=MockArg4()
        ans = coopr.opt.convert_problem( (arg,ProblemFormat.nl,arg), None, [ProblemFormat.nl])
        self.assertNotEqual(re.match(".*tmp.*pyomo.nl$",ans[0][0]), None)
        os.remove(ans[0][0])

    def test_pyomo_nl1(self):
        """ Convert from Pyomo to NL with file"""
        ans = coopr.opt.convert_problem( (currdir+'model.py',ProblemFormat.nl,), None, [ProblemFormat.nl])
        self.assertNotEqual(re.match(".*tmp.*pyomo.nl$",ans[0][0]), None)
        os.remove(ans[0][0])

    def test_error1(self):
        """ No valid problem types """
        try:
            coopr.opt.convert_problem( ("test4.nl","tmp.nl"), ProblemFormat.nl, [])
            self.fail("Expected ConverterError exception")
        except ConverterError:
            err = sys.exc_info()[1]
            pass

    def test_error2(self):
        """ Target problem type is not valid """
        try:
            coopr.opt.convert_problem( ("test4.nl","tmp.nl"), ProblemFormat.nl, [ProblemFormat.mps])
            self.fail("Expected ConverterError exception")
        except ConverterError:
            pass

    def test_error3(self):
        """ Empty argument list """
        try:
            coopr.opt.convert_problem( (), None, [ProblemFormat.mps])
            self.fail("Expected ConverterError exception")
        except ConverterError:
            pass

    def test_error4(self):
        """ Unknown source type """
        try:
            coopr.opt.convert_problem( ("prob.foo",), None, [ProblemFormat.mps])
            self.fail("Expected ConverterError exception")
        except ConverterError:
            pass

    def test_error5(self):
        """ Unknown source type """
        try:
            coopr.opt.convert_problem( ("prob.lp",), ProblemFormat.nl, [ProblemFormat.nl])
            self.fail("Expected ConverterError exception")
        except ConverterError:
            pass

    def test_error6(self):
        """ Cannot use pico_convert with more than one file """
        try:
            ans = coopr.opt.convert_problem( (currdir+"test4.nl","foo"), None, [ProblemFormat.cpxlp])
            self.fail("Expected ConverterError exception")
        except ConverterError:
            pass

    def test_error8(self):
        """ Error when source file cannot be found """
        try:
            ans = coopr.opt.convert_problem( (currdir+"unknown.nl",), None, [ProblemFormat.cpxlp])
            self.fail("Expected ConverterError exception")
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyutilib.services.registered_executable("pico_convert").enabled():
                self.fail("Expected ApplicationError because pico_convert is not available: '%s'" % str(err))
            return
        except ConverterError:
            pass

    def test_error9(self):
        """ The Opt configuration has not been initialized """
        cmd = pyutilib.services.registered_executable("pico_convert")
        if not cmd is None:
            cmd.disable()
        try:
            ans = coopr.opt.convert_problem( (currdir+"test4.nl",), None, [ProblemFormat.cpxlp])
            self.fail("This test didn't fail, but pico_convert should not be defined.")
        except ConverterError:
            pass
        if not cmd is None:
            cmd.enable()

    def test_error10(self):
        """ GLPSOL can only convert file data """
        try:
            arg = MockArg3()
            ans = coopr.opt.convert_problem( (arg,ProblemFormat.cpxlp,arg), None, [ProblemFormat.cpxlp])
            self.fail("This test didn't fail, but glpsol cannot handle objects.")
        except ConverterError:
            pass

    def test_error11(self):
        """ Cannot convert MOD that contains data """
        try:
            ans = coopr.opt.convert_problem( (currdir+"test3.mod",currdir+"test5.dat"), None, [ProblemFormat.cpxlp])
            self.fail("Expected ConverterError exception because we provided a MOD file with a 'data;' declaration")
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyutilib.services.registered_executable("glpsol").enabled():
                self.fail("Expected ApplicationError because glpsol is not available: '%s'" % str(err))
            return
        except ConverterError:
            pass

if __name__ == "__main__":
    unittest.main()
